import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from preliz.distributions import (
    Gamma,
    Hurdle,
    LogNormal,
    NegativeBinomial,
    Normal,
    Poisson,
    Truncated,
)
from preliz.internal.distribution_helper import eps


@pytest.mark.parametrize(
    "dist",
    [
        (Gamma(3, 5)),
        (LogNormal(0, 0.5)),
        (Normal(0, 2)),
        (Poisson(3.5)),
        (NegativeBinomial(3, 5)),
    ],
)
def test_hurdle_vs_truncated(dist):
    if dist.kind == "discrete":
        lower = 1
    else:
        lower = eps

    hurdle_dist = Hurdle(dist, 1)
    base_dist = Truncated(dist, lower=lower, upper=np.inf)

    rng = np.random.default_rng(1)
    actual_rvs = hurdle_dist.rvs(100_000, random_state=rng)
    expected_rvs = base_dist.rvs(100_000, random_state=rng)
    assert_almost_equal(actual_rvs.mean(), expected_rvs.mean(), decimal=2)
    assert_almost_equal(actual_rvs.var(), expected_rvs.var(), decimal=2)

    assert_almost_equal(hurdle_dist.mean(), base_dist.mean(), decimal=2)
    assert_almost_equal(hurdle_dist.var(), base_dist.var(), decimal=3)
    assert_almost_equal(hurdle_dist.std(), base_dist.std(), decimal=4)
    assert_almost_equal(hurdle_dist.entropy(), base_dist.entropy())

    few_rvs = hurdle_dist.rvs(20, random_state=rng)
    assert_almost_equal(hurdle_dist.pdf(few_rvs), base_dist.pdf(few_rvs))
    assert_almost_equal(hurdle_dist.cdf(few_rvs), base_dist.cdf(few_rvs))
    assert_almost_equal(hurdle_dist.logpdf(few_rvs), base_dist.logpdf(few_rvs))
    assert_almost_equal(hurdle_dist._neg_logpdf(few_rvs), base_dist._neg_logpdf(few_rvs))
    x_vals = [-1, 0, 0.25, 0.5, 0.75, 1, 2]
    assert_almost_equal(hurdle_dist.ppf(x_vals), base_dist.ppf(x_vals))


@pytest.mark.parametrize(
    "dist",
    [
        (Gamma(3, 5)),
        (LogNormal(0, 0.5)),
        (Normal(0, 2)),
        (Poisson(3.5)),
        (NegativeBinomial(3, 5)),
    ],
)
def test_hurdle_vs_random(dist):
    hurdle_dist = Hurdle(dist, psi=0.7)

    rng = np.random.default_rng(1)
    rvs = hurdle_dist.rvs(500_000, random_state=rng)

    assert_almost_equal(hurdle_dist.mean(), rvs.mean(), decimal=2)
    assert_almost_equal(hurdle_dist.var(), rvs.var(), decimal=0)
    assert_almost_equal(hurdle_dist.std(), rvs.std(), decimal=0)



import numpy as np

from preliz.distributions.distributions import DistributionTransformer
from preliz.internal.distribution_helper import all_not_none, eps


class Hurdle(DistributionTransformer):
    r"""
    Hurdle distribution.

    This is not a distribution per se, but a modifier of univariate distributions.

    Given a base distribution with parameters $\theta$, cumulative distribution function (CDF),
    and probability density mass/function (PDF). The density of a Hurdle distribution is:

    .. math::

        f(x \mid \psi, \mu) =
            \left\{
                \begin{array}{l}
                (1 - \psi)  \ \text{if } x = 0 \\
                \psi
                \frac{\text{PDF}(x \mid \theta))}
                {1 - \text{CDF}(\epsilon \mid \theta)} \ \text{if } x \neq 0
                \end{array}
            \right.

    where $\psi$ is the expected proportion of the base distribution (0 < $\psi$ < 1), and
    $\epsilon$ is the machine precision for continuous distributions and 0 for discrete ones.

    The following figure shows the difference between a Gamma distribution and a HurdleGamma, with
    the same parameters for the base distribution (Gamma).

    .. plot::
        :context: close-figs

        from preliz import Gamma, Hurdle, style
        style.use('preliz-doc')
        Hurdle(Gamma(mu=2, sigma=1), 0.8).plot_pdf()
        Gamma(mu=2, sigma=1).plot_pdf()

    Parameters
    ----------
    dist: PreliZ distribution
        Univariate PreliZ distribution which will be truncated.
    psi : float
        Expected proportion of the base distribution (0 < psi < 1)
    """

    def __init__(self, dist, psi=None, **kwargs):
        self.dist = dist
        super().__init__()
        self._parametrization(psi, **kwargs)

    def _parametrization(self, psi=None, **kwargs):
        dist_params = []
        if not kwargs:
            if hasattr(self.dist, "params"):
                kwargs = dict(zip(self.dist.param_names, self.dist.params))
            else:
                kwargs = dict(zip(self.dist.param_names, [None] * len(self.dist.param_names)))

        for key, value in kwargs.items():
            dist_params.append(value)
            setattr(self, key, value)

        self.psi = psi
        self.params = (*dist_params, self.psi)
        self.param_names = (*self.dist.param_names, "psi")
        if all_not_none(*dist_params):
            self.dist._parametrization(**kwargs)
            self.is_frozen = True

        self.support = self.dist.support
        self.params_support = (*self.dist.params_support, (0, 1))

    def mean(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        if self.kind == "discrete":
            return np.sum(x_values * pdf)
        else:
            return np.trapz(x_values * pdf, x_values)

    def mode(self):
        if self.kind == "discrete":
            from preliz.internal.optimization import find_discrete_mode

            return find_discrete_mode(self)
        else:
            from preliz.internal.optimization import find_mode

            return find_mode(self)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        if self.kind == "discrete":
            return np.sum((x_values - self.mean()) ** 2 * pdf)
        else:
            return np.trapz((x_values - self.mean()) ** 2 * pdf, x_values)

    def std(self):
        return self.var() ** 0.5

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return self.ppf(random_state.uniform(size=size))

    def pdf(self, x):
        if self.dist == "discrete":
            return np.where(
                x == 0, 1 - self.psi, self.psi * self.dist.pdf(x) / (1 - self.dist.cdf(0))
            )
        else:
            return np.where(
                x == 0, 1 - self.psi, self.psi * self.dist.pdf(x) / (1 - self.dist.cdf(eps))
            )

    def cdf(self, x):
        if self.dist == "discrete":
            return np.where(
                x <= 0, 1 - self.psi, 1 - self.psi * (1 - self.dist.cdf(x)) / (1 - self.dist.cdf(0))
            )
        else:
            return np.where(
                x <= 0,
                1 - self.psi,
                1 - self.psi * (1 - self.dist.cdf(x)) / (1 - self.dist.cdf(eps)),
            )

    def ppf(self, q):
        if self.kind == "discrete":
            lower = 0
        else:
            lower = eps
        q = np.asarray(q)
        psi = self.psi
        lcdf = self.dist.cdf(lower)
        vals = np.zeros_like(q)
        vals[q >= (1 - psi)] = self.dist.ppf(
            lcdf + (q[q >= (1 - psi)] - (1 - psi)) * (1 - lcdf) / psi
        )
        return np.where((q < 0) | (q > 1), np.nan, vals)

    def logpdf(self, x):
        if self.dist == "discrete":
            pdf_values = np.where(
                x == 0,
                np.log(1 - self.psi),
                np.log(self.psi) + self.dist.logpdf(x) - np.log(1 - self.dist.cdf(0)),
            )
        else:
            pdf_values = np.where(
                x == 0,
                np.log(1 - self.psi),
                np.log(self.psi) + self.dist.logpdf(x) - np.log(1 - self.dist.cdf(eps)),
            )
        return pdf_values

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        if self.kind == "discrete":
            return -np.sum(np.exp(logpdf) * logpdf)
        else:
            return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def _neg_logpdf(self, x):
        return -self.logpdf(x).sum()

    def _fit_moments(self, mean, sigma):
        self.dist._fit_moments(mean, sigma)
        self._parametrization(**dict(zip(self.dist.param_names, self.dist.params)))



from preliz.unidimensional.combine import combine
from preliz.unidimensional.combine_roulette import combine_roulette
from preliz.unidimensional.maxent import maxent
from preliz.unidimensional.mle import mle
from preliz.unidimensional.quartile import quartile
from preliz.unidimensional.quartile_int import QuartileInt
from preliz.unidimensional.roulette import Roulette

__all__ = [
    "combine",
    "combine_roulette",
    "maxent",
    "mle",
    "quartile",
    "QuartileInt",
    "Roulette",
]



import pytest

import preliz as pz


@pytest.mark.parametrize(
    "methods, filepath, format_type",
    [
        (None, None, "bibtex"),
        ([pz.Roulette, pz.dirichlet_mode], None, "bibtex"),
        (["all", None, "bibtex"]),
    ],
)
def test_citations(methods, filepath, format_type):
    pz.citations(methods, filepath, format_type)



import numpy as np
from numpy.testing import assert_almost_equal
from scipy import special as sc_special

from preliz.internal import special as pz_special


def test_erf():
    x = np.linspace(-2, 2, 100)
    assert_almost_equal(sc_special.erf(x), pz_special.erf(x))


def test_inv_erf():
    x = np.linspace(-0.9, 0.9, 100)
    assert_almost_equal(sc_special.erfinv(x), pz_special.erfinv(x))


def test_beta():
    a = np.linspace(0, 10, 100)
    b = np.linspace(0, 10, 100)
    assert_almost_equal(sc_special.beta(a, b), pz_special.beta(a, b))


def test_betaln():
    a = np.linspace(0, 10, 100)
    b = np.linspace(0, 10, 100)
    assert_almost_equal(sc_special.betaln(a, b), pz_special.betaln(a, b))


def test_betainc():
    x = np.linspace(0, 1, 100)
    a = np.linspace(0, 10, 100)
    b = np.linspace(0, 10, 100)
    assert_almost_equal(sc_special.betainc(a, b, x), pz_special.betainc(a, b, x))


def test_betaincinv():
    x = np.linspace(0, 1, 100)
    a = np.linspace(0, 10, 100)
    b = np.linspace(0, 10, 100)
    # in scipy < 1.12 this matches to at least 1e-7
    # in scipy >= 1.12 this matches to 1e-5
    assert_almost_equal(sc_special.betaincinv(a, b, x), pz_special.betaincinv(a, b, x))


def test_gammaln():
    x = np.linspace(0.1, 10, 100)
    assert_almost_equal(sc_special.gammaln(x), pz_special.gammaln(x))


def test_gamma():
    x = np.linspace(0.1, 10, 100)
    assert_almost_equal(sc_special.gamma(x), pz_special.gamma(x))


def test_digamma():
    x = np.linspace(0.1, 10, 100)
    assert_almost_equal(sc_special.digamma(x), pz_special.digamma(x))


def test_logit():
    x = np.linspace(-0.1, 1.1, 100)
    assert_almost_equal(sc_special.logit(x), pz_special.logit(x))


def test_expit():
    x = np.linspace(-20, 10, 500)
    assert_almost_equal(sc_special.expit(x), pz_special.expit(x))


def test_xlogy():
    x = np.linspace(0, 10, 10)
    y = np.linspace(0, 10, 10)
    assert_almost_equal(sc_special.xlogy(x, y), pz_special.xlogy(x, y))


def test_xlog1py():
    x = np.linspace(0, 10, 10)
    y = np.linspace(0, 10, 10)
    assert_almost_equal(sc_special.xlog1py(x, y), pz_special.xlog1py(x, y))


def test_xlogx():
    x = np.linspace(0.0, 10, 10)
    assert_almost_equal(pz_special.xlogy(x, x), pz_special.xlogx(x))



import numpy as np
from numpy.testing import assert_array_almost_equal
from scipy.stats import gaussian_kde

from preliz.internal.narviz import _bw_isj, _dct1d, hdi, histogram, kde


def test_hdi():
    normal_sample = np.random.randn(50000)
    interval = hdi(normal_sample)
    assert_array_almost_equal(interval, [-1.8, 1.8], 1)


def test_kde():
    sample = np.random.randn(5000)
    grid, pdf = kde(sample)
    scipy_kde = gaussian_kde(sample)
    scipy_pdf = scipy_kde(grid)
    assert_array_almost_equal(pdf, scipy_pdf, decimal=1)


def test_dct1d():
    data = np.array([1, 2, 3, 4, 5])
    dct_result = _dct1d(data)

    expected_dct_result = np.array([17.0, -9.959, -1.236, -0.898, -3.236])

    assert_array_almost_equal(dct_result, expected_dct_result, decimal=3)


def test_bw_isj():
    sample = np.random.randn(5000)
    grid_counts, _, _ = histogram(sample, bins=256)
    x_range = sample.max() - sample.min()
    bandwidth = _bw_isj(sample, grid_counts=grid_counts, x_range=x_range)

    # Using a simple heuristic for expected bandwidth
    expected_bandwidth = 1.06 * np.std(sample) * len(sample) ** (-1 / 5)

    assert_array_almost_equal(bandwidth, expected_bandwidth, decimal=1)



import numba as nb
import numpy as np
from scipy.special import nbdtrik

from preliz.distributions.distributions import Discrete
from preliz.internal.distribution_helper import all_not_none, any_not_none, eps
from preliz.internal.optimization import find_discrete_mode, optimize_ml, optimize_moments
from preliz.internal.special import betainc, cdf_bounds, gammaln, ppf_bounds_disc, xlogy


class ZeroInflatedNegativeBinomial(Discrete):
    R"""
    Zero-Inflated Negative binomial distribution.

    The Zero-inflated version of the Negative Binomial (NB).
    The NB distribution describes a Poisson random variable
    whose rate parameter is gamma distributed.
    The pmf of this distribution is

    .. math::

       f(x \mid \psi, \mu, \alpha) = \left\{
         \begin{array}{l}
           (1-\psi) + \psi \left (
             \frac{\alpha}{\alpha+\mu}
           \right) ^\alpha, \text{if } x = 0 \\
           \psi \frac{\Gamma(x+\alpha)}{x! \Gamma(\alpha)} \left (
             \frac{\alpha}{\mu+\alpha}
           \right)^\alpha \left(
             \frac{\mu}{\mu+\alpha}
           \right)^x, \text{if } x=1,2,3,\ldots
         \end{array}
       \right.

    .. plot::
        :context: close-figs

        from preliz import ZeroInflatedNegativeBinomial, style
        style.use('preliz-doc')
        psis = [0.7, 0.7]
        mus = [2, 8]
        alphas = [2, 4]
        for psi, mu, alpha in zip(psis, mus, alphas):
            ZeroInflatedNegativeBinomial(psi, mu=mu, alpha=alpha).plot_pdf(support=(0,25))

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\psi\mu`
    Var       .. math::
                  \psi \left(\frac{{\mu^2}}{{\alpha}}\right) +\
                  \psi \mu + \psi \mu^2 - \psi^2 \mu^2
    ========  ==========================

    The zero inflated negative binomial distribution can be parametrized
    either in terms of mu and alpha, or in terms of n and p.
    The link between the parametrizations is given by

    .. math::

        \mu &= \frac{n(1-p)}{p} \\
        \alpha &= n

    Parameters
    ----------
    psi : float
        Expected proportion of NegativeBinomial variates (0 < psi < 1)
    mu : float
        Poisson distribution parameter (mu > 0).
    alpha : float
        Gamma distribution parameter (alpha > 0).
    p : float
        Alternative probability of success in each trial (0 < p < 1).
    n : float
        Alternative number of target success trials (n > 0)
    """

    def __init__(self, psi=None, mu=None, alpha=None, p=None, n=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(psi, mu, alpha, p, n)

    def _parametrization(self, psi=None, mu=None, alpha=None, p=None, n=None):
        if any_not_none(mu, alpha) and any_not_none(p, n):
            raise ValueError(
                "Incompatible parametrization. Either use psi, mu and alpha, or psi, p and n."
            )

        self.psi = psi
        self.param_names = ("psi", "mu", "alpha")
        self.params_support = ((eps, 1 - eps), (eps, np.inf), (eps, np.inf))

        if any_not_none(p, n):
            self.p = p
            self.n = n
            self.param_names = ("psi", "p", "n")
            if all_not_none(p, n):
                mu, alpha = self._from_p_n(p, n)

        self.mu = mu
        self.alpha = alpha
        self.params = (self.psi, self.mu, self.alpha)
        if all_not_none(mu, alpha):
            self._update(psi, mu, alpha)

    def _from_p_n(self, p, n):
        alpha = n
        mu = n * (1 / p - 1)
        return mu, alpha

    def _to_p_n(self, mu, alpha):
        p = alpha / (mu + alpha)
        n = alpha
        return p, n

    def _update(self, psi, mu, alpha):
        self.psi = np.float64(psi)
        self.mu = np.float64(mu)
        self.alpha = np.float64(alpha)
        self.p, self.n = self._to_p_n(self.mu, self.alpha)

        if self.param_names[1] == "mu":
            self.params = (self.psi, self.mu, self.alpha)
        elif self.param_names[1] == "p":
            self.params = (self.psi, self.p, self.n)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.psi, self.n, self.p, self.mu))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.psi, self.n, self.p, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.psi, self.n, self.p, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.psi, self.n, self.p, self.mu)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.psi, self.n, self.p, self.mu)

    def entropy(self):
        x = self.xvals("full", 5000)
        logpdf = self.logpdf(x)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        return self.psi * self.mu

    def mode(self):
        return find_discrete_mode(self)

    def median(self):
        # missing explicit expression
        return self.ppf(0.5)

    def var(self):
        var_nb = self.mu**2 / self.alpha + self.mu
        return self.psi * (var_nb + self.mu**2) - (self.psi * self.mu) ** 2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        # implement skewness
        return np.nan

    def kurtosis(self):
        # implement kurtosis
        return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        zeros = random_state.uniform(size=size) > (1 - self.psi)
        nbinomial = random_state.negative_binomial(self.n, self.p, size=size)
        return zeros * nbinomial

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, psi, n, p, lower, upper):
    nb_prob = betainc(n, x + 1, p)
    prob = (1 - psi) + psi * nb_prob
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
# bdtrik not supported by numba
def nb_ppf(q, psi, n, p, lower, upper):
    nb_vals = np.ceil(nbdtrik(q, n, p))
    x_vals = (1 - psi) + psi * nb_vals
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(y, psi, n, p, mu):
    if y < 0:
        return -np.inf
    elif y == 0:
        return np.log((1 - psi) + psi * (n / (n + mu)) ** n)
    else:
        return (
            np.log(psi)
            + gammaln(y + n)
            - gammaln(n)
            - gammaln(y + 1)
            + xlogy(n, p)
            + xlogy(y, 1 - p)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(y, psi, n, p, mu):
    return -(nb_logpdf(y, psi, n, p, mu)).sum()



try:
    from IPython.display import display
    from ipywidgets import HBox, VBox, interactive_output
except ImportError:
    pass
from preliz.internal.plot_helper import get_textboxes
from preliz.ppls.agnostic import (
    inspect_source,
    parse_function_for_pred_textboxes,
    ppl_plot_decorator,
)


def predictive_explorer(
    fmodel, samples=50, kind_plot="ecdf", references=None, plot_func=None, engine="auto"
):
    """
    Explore how changing parameters in the prior affects the prior predictive distribution.

    Use this function to interactively explore how a prior predictive distribution changes when the
    priors are changed.

    Parameters
    ----------
    fmodel : callable
        A function with a PreliZ model, a PyMC model, or a Bambi model.
        See examples section below for details
    samples : int, optional
        The number of samples to draw from the prior predictive distribution (default is 50).
    kind_plot : str, optional
        The type of plot to display. Defaults to "kde". Options are "hist" (histogram),
        "kde" (kernel density estimate), "ecdf" (empirical cumulative distribution function).
    references : int, float, list, tuple or dictionary
        Value(s) used as reference points representing prior knowledge. For example expected
        values or values that are considered extreme. Use a dictionary for labeled references.
    plot_func : function
        Custom matplotlib code. Defaults to None. ``kind_plot`` and ``references`` are ignored
        if ``plot_func`` is specified.
    engine : str, optional
        Library used to define the fmodel. Either `preliz`, `pymc` or `bambi`. Default is `auto`.
        The function will automatically select the appropriate library to use based on the fmodel
        provided.
    """
    source, signature, engine = inspect_source(fmodel)
    model = parse_function_for_pred_textboxes(source, signature, engine)
    textboxes = get_textboxes(signature, model)
    new_fmodel = ppl_plot_decorator(fmodel, samples, kind_plot, references, plot_func, engine)
    out = interactive_output(new_fmodel, textboxes)
    default_names = [
        "__set_xlim__",
        "__x_min__",
        "__x_max__",
        "__set_ylim__",
        "__y_min__",
        "__y_max__",
        "__resample__",
    ]
    default_controls = [textboxes[name] for name in default_names]
    params_controls = [v for k, v in textboxes.items() if k not in default_names]

    params_plot = VBox(params_controls + [out])

    display(HBox([params_plot, VBox(default_controls)]))



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from scipy.stats import kurtosis, skew

from preliz.distributions import Mixture, Normal, Poisson


@pytest.mark.parametrize(
    "dist0, dist1, weights",
    [
        (Normal(-1.5, 1), Normal(1, 0.5), [0.6, 0.4]),
        (Poisson(4), Poisson(1.5), [0.5, 0.5]),
    ],
)
def test_mixture(dist0, dist1, weights):
    mix_dist = Mixture([dist0, dist1], weights=weights)
    s_s = (np.array(weights) * 10000).astype(int)
    mix_samples = np.concatenate([dist0.rvs(s_s[0]), dist1.rvs(s_s[1])])

    assert_almost_equal(mix_dist.mean(), mix_samples.mean(), decimal=1)
    assert_almost_equal(mix_dist.var(), mix_samples.var(), decimal=0)
    assert_almost_equal(skew(mix_samples), mix_dist.skewness(), decimal=1)
    assert_almost_equal(kurtosis(mix_samples), mix_dist.kurtosis(), decimal=0)
    assert_almost_equal(mix_dist.median(), np.median(mix_samples), decimal=1)
    assert_almost_equal(mix_dist.std(), mix_samples.std(), decimal=1)
    assert_almost_equal(mix_dist.rvs(10000).mean(), mix_samples.mean(), decimal=1)

    # x_vals = cen_dist.rvs(1000000, random_state=1)
    # assert_almost_equal(np.mean(x_vals == lower), dist.cdf(lower), decimal=2)
    # if dist.kind == "discrete":
    #     assert_almost_equal(np.mean(x_vals == upper), 1 - dist.cdf(upper - 1), decimal=2)
    # else:
    #     assert_almost_equal(np.mean(x_vals == upper), 1 - dist.cdf(upper), decimal=2)

    # x_inside = x_vals[(x_vals > lower) & (x_vals < upper)]
    # assert_almost_equal(dist.logpdf(x_inside), cen_dist.logpdf(x_inside))
    # assert_almost_equal(dist.cdf(x_inside), cen_dist.cdf(x_inside))

    # assert_almost_equal(dist.cdf(x_inside), cen_dist.cdf(x_inside))

    # assert_almost_equal(cen_dist.median(), dist.median())
    # assert_almost_equal(x_vals.mean(), cen_dist.mean(), decimal=1)
    # assert_almost_equal(x_vals.var(), cen_dist.var(), decimal=1)
    # assert_almost_equal(skew(x_vals), cen_dist.skewness(), decimal=0)
    # assert_almost_equal(kurtosis(x_vals), cen_dist.kurtosis(), decimal=0)

    # actual_mean = dist.mean()
    # expected_mean = cen_dist_inf.mean()
    # assert_almost_equal(actual_mean, expected_mean, decimal=2)

    # actual_var = dist.var()
    # expected_var = cen_dist_inf.var()
    # assert_almost_equal(actual_var, expected_var, decimal=2)

    # actual_entropy = dist.entropy()
    # expected_entropy = cen_dist_inf.entropy()
    # assert_almost_equal(actual_entropy, expected_entropy, decimal=1)

    # c_l, c_u = cen_dist.hdi()
    # d_l, d_u = dist.hdi()
    # assert c_l >= d_l
    # assert c_u <= d_u
    # assert_almost_equal(cen_dist_inf.hdi(), dist.hdi())



import numba as nb
import numpy as np
from scipy.special import zeta

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml
from preliz.internal.special import cdf_bounds, ppf_bounds_cont


class Gumbel(Continuous):
    r"""
    Gumbel distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu, \beta) = \frac{1}{\beta}e^{-(z + e^{-z})}

    where

    .. math::

        z = \frac{x - \mu}{\beta}

    .. plot::
        :context: close-figs


        from preliz import Gumbel, style
        style.use('preliz-doc')
        mus = [0., 4., -1.]
        betas = [1., 2., 4.]
        for mu, beta in zip(mus, betas):
            Gumbel(mu, beta).plot_pdf(support=(-10,20))

    ========  ==========================================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu + \beta\gamma`, where :math:`\gamma` is the Euler-Mascheroni constant
    Variance  :math:`\frac{\pi^2}{6} \beta^2`
    ========  ==========================================

    Parameters
    ----------
    mu : float
        Location parameter.
    beta : float
        Scale parameter (beta > 0).
    """

    def __init__(self, mu=None, beta=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, beta)

    def _parametrization(self, mu=None, beta=None):
        self.mu = mu
        self.beta = beta
        self.params = (self.mu, self.beta)
        self.param_names = ("mu", "beta")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        if all_not_none(self.mu, self.beta):
            self._update(self.mu, self.beta)

    def _update(self, mu, beta):
        self.mu = np.float64(mu)
        self.beta = np.float64(beta)
        self.params = (self.mu, self.beta)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.beta, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.beta, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.beta)

    def entropy(self):
        return nb_entropy(self.beta)

    def mean(self):
        return self.mu + self.beta * np.euler_gamma

    def mode(self):
        return self.mu

    def median(self):
        return self.mu - self.beta * np.log(np.log(2))

    def var(self):
        return np.pi**2 / 6 * self.beta**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 12 * 6**0.5 * zeta(3) / np.pi**3

    def kurtosis(self):
        return 12 / 5

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return nb_ppf(random_state.uniform(size=size), self.mu, self.beta, -np.inf, np.inf)

    def _fit_moments(self, mean, sigma):
        beta = sigma / np.pi * 6**0.5
        mu = mean - beta * np.euler_gamma
        self._update(mu, beta)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, mu, beta, lower, upper):
    prob = np.exp(-np.exp(-(x - mu) / beta))
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, mu, beta, lower, upper):
    x_val = mu - beta * np.log(-np.log(q))
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(beta):
    return np.log(beta) + 1 + np.euler_gamma


@nb.njit(cache=True)
def nb_logpdf(x, mu, beta):
    zval = (x - mu) / beta
    return -(zval + np.exp(-zval) + np.log(beta))


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, beta):
    return -(nb_logpdf(x, mu, beta)).sum()



import numba as nb
import numpy as np


@nb.vectorize(nopython=True, cache=True)
def erf(x):
    """
    Error function.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    sign = 1
    if x < 0:
        sign = -1

    x = np.abs(x)

    t = 1.0 / (1.0 + 0.3275911 * x)
    y = 1.0 - (
        ((((1.061405429 * t + -1.453152027) * t) + 1.421413741) * t + -0.284496736) * t
        + 0.254829592
    ) * t * np.exp(-x * x)

    return sign * y


ERFC_COF = np.array(
    [
        -2.8e-17,
        1.21e-16,
        -9.4e-17,
        -1.523e-15,
        7.106e-15,
        3.81e-16,
        -1.12708e-13,
        3.13092e-13,
        8.94487e-13,
        -6.886027e-12,
        2.394038e-12,
        9.6467911e-11,
        -2.27365122e-10,
        -9.91364156e-10,
        5.059343495e-9,
        6.529054439e-9,
        -8.5238095915e-8,
        1.5626441722e-8,
        1.303655835580e-6,
        -1.624290004647e-6,
        -2.0278578112534e-5,
        4.2523324806907e-5,
        3.66839497852761e-4,
        -9.46595344482036e-4,
        -9.561514786808631e-3,
        1.9476473204185836e-2,
        6.4196979235649026e-1,
        -1.3026537197817094,
    ]
)
ERFC_COF_LAST = ERFC_COF[-1]


@nb.vectorize(nopython=True, cache=True)
def erfccheb(y):
    d = 0.0
    dd = 0.0
    temp = 0.0
    t = 2.0 / (2.0 + y)
    ty = 4.0 * t - 2.0

    for i in range(len(ERFC_COF) - 1):
        temp = d
        d = ty * d - dd + ERFC_COF[i]
        dd = temp

    return t * np.exp(-y * y + 0.5 * (ERFC_COF_LAST + ty * d) - dd)


@nb.vectorize(nopython=True, cache=True)
def erfc(x):
    return erfccheb(x) if x >= 0.0 else 2.0 - erfccheb(-x)


@nb.vectorize(nopython=True, cache=True)
def erfcinv(p):
    """
    Invert complementary error function.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    if p < 0.0 or p > 2.0:
        raise ValueError("Argument must be between 0 and 2")
    elif p == 0.0:
        return np.inf
    elif p == 2.0:
        return -np.inf
    else:
        pp = p if p < 1.0 else 2.0 - p
        t = np.sqrt(-2.0 * np.log(pp / 2.0))
        x = -0.70711 * ((2.30753 + t * 0.27061) / (1.0 + t * (0.99229 + t * 0.04481)) - t)

        err1 = erfc(x) - pp
        x += err1 / (1.12837916709551257 * np.exp(-x * x) - x * err1)
        err2 = erfc(x) - pp
        x += err2 / (1.12837916709551257 * np.exp(-x * x) - x * err2)

        return x if p < 1.0 else -x


@nb.njit(cache=True)
def erfinv(p):
    return -erfcinv(p + 1)


@nb.njit(cache=True)
def erfcx(x):
    return np.exp(x**2) * erfc(x)


@nb.vectorize(nopython=True, cache=True)
def beta(a, b):
    if a < 0 or b < 0:
        raise ValueError("Arguments must be positive.")
    elif a == 0 and b == 0:
        return np.inf
    elif a == 0 or b == 0:
        return np.inf

    return np.exp(betaln(a, b))


@nb.vectorize(nopython=True, cache=True)
def betaln(a, b):
    if a < 0 or b < 0:
        raise ValueError("Arguments must be positive.")
    elif a == 0 and b == 0:
        return np.inf
    elif a == 0 or b == 0:
        return np.inf

    return gammaln(a) + gammaln(b) - gammaln(a + b)


@nb.vectorize(nopython=True, cache=True)
def betacf(x, a, b):
    """
    Evaluate the continued fraction for incomplete beta function by modified Lentz's method.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    fpmin = 1e-30
    c = 1
    qab = a + b
    qap = a + 1
    qam = a - 1
    d = 1 - qab * x / qap

    if abs(d) < fpmin:
        d = fpmin

    d = 1 / d
    h = d

    for m in range(1, 101):
        m2 = 2 * m
        aa = m * (b - m) * x / ((qam + m2) * (a + m2))
        d = 1 + aa * d

        if abs(d) < fpmin:
            d = fpmin

        c = 1 + aa / c

        if abs(c) < fpmin:
            c = fpmin

        d = 1 / d
        h *= d * c
        aa = -(a + m) * (qab + m) * x / ((a + m2) * (qap + m2))
        d = 1 + aa * d

        if abs(d) < fpmin:
            d = fpmin

        c = 1 + aa / c

        if abs(c) < fpmin:
            c = fpmin

        d = 1 / d
        del_ = d * c
        h *= del_

        if abs(del_ - 1.0) < 3e-7:
            break

    return h


@nb.vectorize(nopython=True, cache=True)
def betainc(a, b, x):
    """
    Return the regularized incomplete beta function.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    if x < 0 or x > 1:
        raise ValueError("Third argument must be between 0 and 1.")
    elif a == 1 and b == 1:
        return x
    elif a == 0 or b == 0:
        return np.nan
    elif x == 0:
        return 0
    elif x == 1:
        return 1
    else:
        bt = np.exp(gammaln(a + b) - gammaln(a) - gammaln(b) + a * np.log(x) + b * np.log1p(-x))

        if x < (a + 1) / (a + b + 2):
            return bt * betacf(x, a, b) / a
        else:
            return 1 - bt * betacf(1 - x, b, a) / b


@nb.vectorize(nopython=True, cache=True)
def betainc_un(a, b, x):
    """
    Return the "unregularized" incomplete beta function.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    return betainc(a, b, x) * beta(a, b)


@nb.vectorize(nopython=True, cache=True)
def betaincinv(a, b, p):
    """
    Return the inverse of incomplete beta function.

    Note:
    -----
    Adapted from Andreas Madsen's mathfn library
    """
    if p < 0 or p > 1:
        raise ValueError("Third argument must be between 0 and 1.")
    elif a == 1 and b == 1:
        return p
    elif p == 1:
        return 1
    elif p == 0:
        return np.nan
    else:
        EPS = 1e-8
        a1 = a - 1
        b1 = b - 1
        j = 0

        if a >= 1 and b >= 1:
            if p < 0.5:
                pp = p
            else:
                pp = 1 - p

            t = np.sqrt(-2 * np.log(pp))
            x = (2.30753 + t * 0.27061) / (1 + t * (0.99229 + t * 0.04481)) - t
            if p < 0.5:
                x = -x

            al = (x**2 - 3) / 6
            h = 2 / (1 / (2 * a - 1) + 1 / (2 * b - 1))
            w = (x * np.sqrt(al + h) / h) - ((1 / (2 * b - 1)) - (1 / (2 * a - 1))) * (
                al + 5 / 6 - 2 / (3 * h)
            )
            x = a / (a + b * np.exp(2 * w))
        else:
            lna = np.log(a / (a + b))
            lnb = np.log(b / (a + b))
            t = np.exp(a * lna) / a
            u = np.exp(b * lnb) / b
            w = t + u

            if p < t / w:
                x = (a * w * p) ** (1 / a)
            else:
                x = 1 - ((b * w * (1 - p)) ** (1 / b))

        afac = -betaln(a, b)

        for j in range(10):
            if x in (0, 1):
                return x

            err = betainc(a, b, x) - p
            t = np.exp(a1 * np.log(x) + b1 * np.log1p(-x) + afac)
            u = err / t
            t = u / (1 - 0.5 * min(1, u * (a1 / x - b1 / (1 - x))))
            x -= t

            if x <= 0:
                x = 0.5 * (x + t)
            if x >= 1:
                x = 0.5 * (x + t + 1)

            if abs(t) < EPS * x and j > 0:
                break

        return x


@nb.njit(cache=True)
def garcia_approximation(mean, sigma):
    """
    Approximate method of moments for Weibull distribution.

    The approximation is good for values of alpha larger than 0.83.

    Oscar Garcia. Simplified method-of-moments estimation for the Weibull distribution. 1981.
    New Zealand Journal of Forestry Science 11:304-306
    https://www.scionresearch.com/__data/assets/pdf_file/0010/36874/NZJFS1131981GARCIA304-306.pdf
    """
    ks = [-0.221016417, 0.010060668, 0.117358987, -0.050999126]
    z = sigma / mean

    poly = 0
    for idx, k in enumerate(ks):
        poly += k * z**idx

    alpha_p = 1 / (z * (1 + (1 - z) ** 2 * poly))
    beta_p = 1 / (gamma(1 + 1 / alpha_p) / (mean))
    return alpha_p, beta_p


@nb.vectorize(nopython=True, cache=True)
def half_erf(x):
    """
    Error function for values of x >= 0, return 0 otherwise.

    Equations 7.1.27 from Abramowitz and Stegun
    Handbook of Mathematical Functions with Formulas, Graphs, and Mathematical Tables
    """
    if x <= 0:
        return 0

    t = 1.0 / (
        1.0
        + 0.0705230784 * x
        + 0.0422820123 * x**2
        + 0.0092705272 * x**3
        + 0.0001520143 * x**4
        + 0.0002765672 * x**5
        + 0.0000430638 * x**6
    )
    approx = 1 - t**16

    return approx


@nb.vectorize(nopython=True, cache=True)
def digamma(x):
    """Faster digamma function assumes x > 0."""
    r = 0
    while x <= 5:
        r -= 1 / x
        x += 1
    f = 1 / (x * x)
    t = f * (
        -1 / 12.0
        + f
        * (
            1 / 120.0
            + f
            * (
                -1 / 252.0
                + f
                * (
                    1 / 240.0
                    + f * (-1 / 132.0 + f * (691 / 32760.0 + f * (-1 / 12.0 + f * 3617 / 8160.0)))
                )
            )
        )
    )

    return r + np.log(x) - 0.5 / x + t


@nb.njit(cache=True)
def gamma(z):
    p = [
        676.5203681218851,
        -1259.1392167224028,
        771.32342877765313,
        -176.61502916214059,
        12.507343278686905,
        -0.13857109526572012,
        9.9843695780195716e-6,
        1.5056327351493116e-7,
    ]

    z = z - 1
    x = np.ones_like(z) * 0.99999999999980993
    for i in range(8):
        x = x + p[i] / (z + i + 1)

    t = z + 7.5
    return np.sqrt(2 * np.pi) * t ** (z + 0.5) * np.exp(-t) * x


@nb.njit(cache=True)
def gammaln(x):
    cof = [
        76.18009172947146,
        -86.50532032941677,
        24.01409824083091,
        -1.231739572450155,
        0.1208650973866179e-2,
        -0.5395239384953e-5,
    ]
    stp = 2.5066282746310005
    fpf = 5.5
    x = x - 1.0
    tmp = x + fpf
    tmp = (x + 0.5) * np.log(tmp) - tmp
    ser = np.ones_like(x) * 1.000000000190015
    for j in range(6):
        x = x + 1
        ser = ser + cof[j] / x
    return tmp + np.log(stp * ser)


@nb.vectorize(nopython=True, cache=True)
def logit(x):
    if x == 0:
        return -np.inf
    elif x == 1:
        return np.inf
    if x < 0 or x > 1:
        return np.nan
    else:
        return np.log(x / (1 - x))


@nb.vectorize(nopython=True, cache=True)
def expit(x):
    if x >= 0:
        return 1 / (1 + np.exp(-x))
    else:
        return np.exp(x) / (1 + np.exp(x))


@nb.vectorize(nopython=True, cache=True)
def xlogy(x, y):
    if x == 0:
        return 0.0
    if y == 0:
        return -np.inf
    else:
        return x * np.log(y)


@nb.vectorize(nopython=True, cache=True)
def xlog1py(x, y):
    if x == 0:
        return 0.0
    else:
        return x * np.log1p(y)


@nb.vectorize(nopython=True, cache=True)
def xlogx(x):
    if x == 0:
        return 0.0
    else:
        return x * np.log(x)


@nb.vectorize(nopython=True, cache=True)
def xprody(x, y):
    if np.isinf(x):
        return 0
    if np.isinf(y):
        return 0
    else:
        return x * y


@nb.vectorize(nopython=True, cache=True)
def norm_logcdf(x):
    t = x * 0.7071067811865476
    if x < -1.0:
        return np.log(erfcx(-t) / 2) - t * t
    else:
        return np.log1p(-erfc(t) / 2)


@nb.vectorize(nopython=True, cache=True)
def cdf_bounds(prob, x, lower, upper):
    if x < lower:
        return 0
    elif x >= upper:
        return 1
    else:
        return prob


@nb.vectorize(nopython=True, cache=True)
def ppf_bounds_disc(x_val, q, lower, upper):
    if q < 0:
        return np.nan
    elif q > 1:
        return np.nan
    elif q == 0:
        return lower - 1
    elif q == 1:
        return upper
    else:
        return x_val


@nb.vectorize(nopython=True, cache=True)
def ppf_bounds_cont(x_val, q, lower, upper):
    if q < 0:
        return np.nan
    elif q > 1:
        return np.nan
    elif q == 0:
        return lower
    elif q == 1:
        return upper
    else:
        return x_val


@nb.njit(cache=True)
def mean_and_std(data):
    n = len(data)
    mean = np.sum(data) / n

    sum_sq_diff = 0
    for x in data:
        sum_sq_diff += (x - mean) ** 2

    std = (sum_sq_diff / n) ** 0.5

    return mean, std


@nb.njit(cache=True)
def mean_sample(sample):
    return np.mean(sample)



import numba as nb
import numpy as np

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps, from_precision, to_precision
from preliz.internal.optimization import optimize_ml
from preliz.internal.special import (
    beta,
    betainc,
    betaincinv,
    cdf_bounds,
    digamma,
    gamma,
    gammaln,
    ppf_bounds_cont,
)


class HalfStudentT(Continuous):
    r"""
    HalfStudentT Distribution.

    The pdf of this distribution is

    .. math::

        f(x \mid \sigma,\nu) =
            \frac{2\;\Gamma\left(\frac{\nu+1}{2}\right)}
            {\Gamma\left(\frac{\nu}{2}\right)\sqrt{\nu\pi\sigma^2}}
            \left(1+\frac{1}{\nu}\frac{x^2}{\sigma^2}\right)^{-\frac{\nu+1}{2}}

    .. plot::
        :context: close-figs

        from preliz import HalfStudentT, style
        style.use('preliz-doc')
        sigmas = [1., 2., 2.]
        nus = [3, 3., 10.]
        for sigma, nu in zip(sigmas, nus):
            HalfStudentT(nu, sigma).plot_pdf(support=(0,10))

    ========  ==========================================
    Support   :math:`x \in [0, \infty)`
    Mean      .. math::
                  2\sigma\sqrt{\frac{\nu}{\pi}}\
                  \frac{\Gamma\left(\frac{\nu+1}{2}\right)}
                  {\Gamma\left(\frac{\nu}{2}\right)(\nu-1)}\, \text{for } \nu > 2
    Variance  .. math::
                  \sigma^2\left(\frac{\nu}{\nu - 2}-\
                  \frac{4\nu}{\pi(\nu-1)^2}\left(\frac{\Gamma\left(\frac{\nu+1}{2}\right)}
                  {\Gamma\left(\frac{\nu}{2}\right)}\right)^2\right) \text{for } \nu > 2\, \infty\
                  \text{for } 1 < \nu \le 2\, \text{otherwise undefined}
    ========  ==========================================

    HalfStudentT distribution has 2 alternative parameterizations. In terms of nu and
    sigma (standard deviation as nu increases) or nu and lam (precision as nu increases).

    The link between the 2 alternatives is given by

    .. math::

        \lambda = \frac{1}{\sigma^2}

    Parameters
    ----------
    nu : float
        Degrees of freedom, also known as normality parameter (nu > 0).
    sigma : float
        Scale parameter (sigma > 0). Converges to the standard deviation as nu
        increases.
    lam : float
        Scale parameter (lam > 0). Converges to the precision as nu increases.
    """

    def __init__(self, nu=None, sigma=None, lam=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(nu, sigma, lam)

    def _parametrization(self, nu=None, sigma=None, lam=None):
        if all_not_none(sigma, lam):
            raise ValueError(
                "Incompatible parametrization. Either use nu and sigma, or nu and lam."
            )

        self.param_names = ("nu", "sigma")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if lam is not None:
            self.lam = lam
            sigma = from_precision(lam)
            self.param_names = ("nu", "lam")

        self.nu = nu
        self.sigma = sigma
        if all_not_none(self.nu, self.sigma):
            self._update(self.nu, self.sigma)

    def _update(self, nu, sigma):
        self.nu = np.float64(nu)
        self.sigma = np.float64(sigma)
        self.lam = to_precision(self.sigma)

        if self.param_names[1] == "sigma":
            self.params = (self.nu, self.sigma)
        elif self.param_names[1] == "lam":
            self.params = (self.nu, self.lam)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.nu, self.sigma))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.nu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.nu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.nu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.nu, self.sigma)

    def entropy(self):
        return nb_entropy(self.nu, self.sigma)

    def mean(self):
        if self.nu > 1:
            gamma0 = gamma((self.nu + 1) / 2)
            gamma1 = gamma(self.nu / 2)
            if np.isfinite(gamma0) and np.isfinite(gamma1):
                mean = (
                    2 * self.sigma * (self.nu / np.pi) ** 0.5 * (gamma0 / (gamma1 * (self.nu - 1)))
                )
            else:
                mean = self.sigma * (2 / np.pi) ** 0.5
        return mean

    def mode(self):
        return np.zeros_like(self.sigma)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        gamma0 = gamma((self.nu + 1) / 2)
        gamma1 = gamma(self.nu / 2)
        if self.nu > 2:
            if np.isfinite(gamma0) and np.isfinite(gamma1):
                var = self.sigma**2 * (
                    (self.nu / (self.nu - 2))
                    - ((4 * self.nu) / (np.pi * (self.nu - 1) ** 2)) * (gamma0 / gamma1) ** 2
                )
            else:
                # assume nu is large enough that the std of the halfnormal is a good approximation
                var = self.sigma**2 * (1 - 2.0 / np.pi)
        return var

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return NotImplemented

    def kurtosis(self):
        return NotImplemented

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return np.abs(random_state.standard_t(self.nu, size) * self.sigma)

    def _fit_moments(self, mean, sigma):
        # if nu is smaller than 2 the variance is not defined,
        # so if that happens we use 2.1 as an approximation
        nu = self.nu
        if nu is None:
            nu = 100
        elif nu <= 2:
            nu = 2.1

        gamma0 = gamma((nu + 1) / 2)
        gamma1 = gamma(nu / 2)
        if np.isfinite(gamma0) and np.isfinite(gamma1):
            sigma = (
                sigma**2
                / ((nu / (nu - 2)) - ((4 * nu) / (np.pi * (nu - 1) ** 2)) * (gamma0 / gamma1) ** 2)
            ) ** 0.5
        else:
            # we assume a Gaussian for large nu
            sigma = sigma / (1 - 2 / np.pi) ** 0.5
        self._update(nu, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, nu, sigma):
    x = x / sigma
    factor = 0.5 * betainc(0.5 * nu, 0.5, nu / (x**2 + nu))
    return cdf_bounds(np.where(x < 0, factor, 1 - factor) * 2 - 1, x, 0, np.inf)


@nb.njit(cache=True)
def nb_ppf(p, nu, sigma):
    p_factor = (p + 1) / 2
    inv_factor = np.where(
        p_factor < 0.5,
        betaincinv(0.5 * nu, 0.5, 2 * p_factor),
        np.sqrt(nu / betaincinv(0.5 * nu, 0.5, 2 - 2 * p_factor) - nu),
    )
    return ppf_bounds_cont(inv_factor * sigma, p, 0, np.inf)


@nb.njit(cache=True)
def nb_entropy(nu, sigma):
    return (
        np.log(sigma)
        + 0.5 * (nu + 1) * (digamma(0.5 * (nu + 1)) - digamma(0.5 * nu))
        + np.log(np.sqrt(nu) * beta(0.5 * nu, 0.5))
        - np.log(2)
    )


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, nu, sigma):
    if x < 0:
        return -np.inf
    else:
        return (
            gammaln((nu + 1) / 2)
            - gammaln(nu / 2)
            - 0.5 * np.log(nu * np.pi * sigma**2)
            - 0.5 * (nu + 1) * np.log(1 + (x / sigma) ** 2 / nu)
            + np.log(2)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, nu, sigma):
    return -(nb_logpdf(x, nu, sigma)).sum()



import pandas as pd

from preliz.distributions import Gamma, LogNormal, Normal
from preliz.ppls.agnostic import posterior_to_prior

SEED = 2945

data = Normal(0, 1).rvs(200, random_state=SEED)

try:
    import pymc as pm

    with pm.Model() as model:
        a = pm.Normal("a", mu=0, sigma=1)
        b = pm.HalfNormal("b", sigma=[1, 1], shape=2)
        y = pm.Normal("y", mu=a, sigma=b[0], observed=data)
        idata = pm.sample(tune=200, draws=500, random_seed=SEED)
except ImportError:
    pass


def test_p2p_pymc():
    posterior_to_prior(model, idata)
    assert 'Gamma\x1b[0m("b", alpha=' in posterior_to_prior(model, idata, new_families="auto")
    posterior_to_prior(model, idata, new_families=[LogNormal()])
    assert 'Gamma\x1b[0m("b", mu=' in posterior_to_prior(
        model, idata, new_families={"b": [Gamma(mu=0)]}
    )


try:
    import bambi as bmb

    bmb_data = pd.DataFrame(
        {
            "y": Normal(0, 1).rvs(117, random_state=SEED + 1),
            "x": Normal(0, 1).rvs(117, random_state=SEED + 2),
            "x1": Normal(0, 1).rvs(117, random_state=SEED + 3),
        }
    )
    bmb_prior = {"Intercept": bmb.Prior("Normal", mu=0, sigma=1)}
    bmb_model = bmb.Model("y ~ x + x1", bmb_data, priors=bmb_prior)
    bmb_idata = bmb_model.fit(tune=200, draws=200, random_seed=SEED)
except ImportError:
    pass


def test_p2p_bambi():
    posterior_to_prior(bmb_model, bmb_idata)
    posterior_to_prior(bmb_model, bmb_idata, new_families="auto")
    posterior_to_prior(bmb_model, bmb_idata, new_families=[LogNormal()])
    assert 'Normal\x1b[0m", mu=' in posterior_to_prior(
        bmb_model, bmb_idata, new_families={"Intercept": [Normal(mu=1, sigma=1)]}
    )



import numpy as np
import pytest

from preliz import dirichlet_mode


def test_dirichlet_mode():
    _, dist = dirichlet_mode([0.22, 0.22, 0.32, 0.22], 0.99, bound=0.02)
    for alpha, expected in zip(dist.alpha, [675, 675, 981.37, 675.0]):
        assert np.isclose(alpha, expected, atol=0.01)


def test_invalid_mass():
    with pytest.raises(ValueError):
        dirichlet_mode([0.22, 0.22, 0.32, 0.22], 1.1, bound=0.02)


def test_invalid_mode():
    with pytest.raises(ValueError):
        dirichlet_mode([-0.2, 0.22, 0.32, 1.22], 0.99, bound=0.02)


def test_plot_beta_mode():
    _, _ = dirichlet_mode(
        mode=[0.22, 0.22, 0.32, 0.22],
        mass=0.99,
        bound=0.02,
        plot=True,
        plot_kwargs={"pointinterval": True},
    )



import numba as nb
import numpy as np
from scipy.special import owens_t
from scipy.stats import skew

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps, from_precision, to_precision
from preliz.internal.optimization import find_ppf, optimize_ml, optimize_moments
from preliz.internal.special import erf, norm_logcdf


class SkewNormal(Continuous):
    r"""
    SkewNormal distribution.

    The pdf of this distribution is

    .. math::

        f(x \mid \mu, \tau, \alpha) =
        2 \Phi((x-\mu)\sqrt{\tau}\alpha) \phi(x,\mu,\tau)

    .. plot::
        :context: close-figs


        from preliz import SkewNormal, style
        style.use('preliz-doc')
        for alpha in [-6, 0, 6]:
            SkewNormal(mu=0, sigma=1, alpha=alpha).plot_pdf()

    ========  ==========================================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu + \sigma \sqrt{\frac{2}{\pi}} \frac {\alpha }{{\sqrt {1+\alpha ^{2}}}}`
    Variance  :math:`\sigma^2 \left(  1-\frac{2\alpha^2}{(\alpha^2+1) \pi} \right)`
    ========  ==========================================

    SkewNormal distribution has 2 alternative parameterizations. In terms of mu, sigma (standard
    deviation) and alpha, or mu, tau (precision) and alpha.

    The link between the 2 alternatives is given by

    .. math::

        \tau = \frac{1}{\sigma^2}

    Parameters
    ----------
    mu : float
        Location parameter.
    sigma : float
        Scale parameter (sigma > 0).
    alpha : float
        Skewness parameter.
    tau : float
        Precision (tau > 0).

    Notes
    -----
    When alpha=0 we recover the Normal distribution and mu becomes the mean,
    and sigma the standard deviation. In the limit of alpha approaching
    plus/minus infinite we get a half-normal distribution.
    """

    def __init__(self, mu=None, sigma=None, alpha=None, tau=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, sigma, alpha, tau)

    def _parametrization(self, mu=None, sigma=None, alpha=None, tau=None):
        if all_not_none(sigma, tau):
            raise ValueError(
                "Incompatible parametrization. Either use mu, sigma and alpha,"
                " or mu, tau and alpha."
            )

        self.param_names = ("mu", "sigma", "alpha")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf), (-np.inf, np.inf))

        if tau is not None:
            self.tau = tau
            sigma = from_precision(tau)
            self.param_names = ("mu", "tau", "alpha")

        self.mu = mu
        self.sigma = sigma
        self.alpha = alpha
        if all_not_none(self.mu, self.sigma, self.alpha):
            self._update(self.mu, self.sigma, self.alpha)

    def _update(self, mu, sigma, alpha):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.alpha = np.float64(alpha)
        self.tau = to_precision(sigma)

        if self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma, self.alpha)
        elif self.param_names[1] == "tau":
            self.params = (self.mu, self.tau, self.alpha)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma, self.alpha)

    def ppf(self, q):
        q = np.asarray(q)
        return find_ppf(self, q)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma, self.alpha)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma, self.alpha)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return self.mu + self.sigma * np.sqrt(2 / np.pi) * self.alpha / np.sqrt(1 + self.alpha**2)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        delta = self.alpha / (1 + self.alpha**2) ** 0.5
        return self.sigma**2 * (1 - 2 * delta**2 / np.pi)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        delta = self.alpha / (1 + self.alpha**2) ** 0.5
        mean_z = (2 / np.pi) ** 0.5 * delta
        return ((4 - np.pi) / 2) * (mean_z**3 / (1 - mean_z**2) ** (3 / 2))

    def kurtosis(self):
        delta = self.alpha / (1 + self.alpha**2) ** 0.5
        return (
            2
            * (np.pi - 3)
            * ((delta * np.sqrt(2 / np.pi)) ** 4 / (1 - 2 * (delta**2) / np.pi) ** 2)
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        u_0 = random_state.normal(size=size)
        v = random_state.normal(size=size)
        d = self.alpha / np.sqrt(1 + self.alpha**2)
        u_1 = d * u_0 + v * np.sqrt(1 - d**2)
        return np.sign(u_0) * u_1 * self.sigma + self.mu

    def _fit_moments(self, mean, sigma):
        if self.alpha is None:
            self.alpha = 0
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        skewness = skew(sample)
        self.alpha = skewness / (1 - skewness**2) ** 0.5
        optimize_ml(self, sample)


def nb_cdf(x, mu, sigma, alpha):
    return 0.5 * (1 + erf((x - mu) / (sigma * 2**0.5))) - 2 * owens_t((x - mu) / sigma, alpha)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, sigma, alpha):
    if x == np.inf:
        return -np.inf
    elif x == -np.inf:
        return -np.inf
    else:
        z_val = (x - mu) / sigma
        return (
            np.log(2)
            - np.log(sigma)
            - z_val**2 / 2.0
            - np.log((2 * np.pi) ** 0.5)
            + norm_logcdf(alpha * z_val)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma, alpha):
    return -(nb_logpdf(x, mu, sigma, alpha)).sum()



from test_helper import run_notebook

from preliz.internal.plot_helper import check_inside_notebook


def test_check_inside_notebook_not(capsys):
    check_inside_notebook()
    captured = capsys.readouterr()
    assert "RuntimeError" in captured.out


def test_check_inside_notebook_yes():
    run_notebook("check_inside_notebook.ipynb")



import warnings

import numpy as np

from preliz.distributions.normal import Normal
from preliz.internal.distribution_helper import valid_distribution
from preliz.internal.optimization import get_fixed_params, optimize_quartile, relative_error
from preliz.internal.rcparams import rcParams


def quartile(
    distribution=None,
    q1=-1,
    q2=0,
    q3=1,
    plot=None,
    plot_kwargs=None,
    ax=None,
):
    """
    Find the distribution with the specified quartiles.

    Parameters
    ----------
    distribution : PreliZ distribution
        Instance of a PreliZ distribution
    q1 : float
        First quartile, i.e 0.25 of the mass is below this point.
    q2 : float
        Second quartile, i.e 0.50 of the mass is below this point. This is also know
        as the median.
    q3 : float
        Third quartile, i.e 0.75 of the mass is below this point.
    plot : bool
        Whether to plot the distribution, and lower and upper bounds. Defaults to None,
        which results in the value of rcParams["plots.show_plot"] being used.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()`` of ``distribution``.
    ax : matplotlib axes

    Returns
    -------
    dict: dict with the parameters of the distribution
    axes: matplotlib axes (only if `plot=True`)

    Notes
    -----
    After calling this function the attribute `opt` of the distribution will be updated with the
    OptimizeResult object from the optimization step.

    See Also
    --------
    maxent : Find the maximum entropy distribution with a given mass inside a user defined interval.

    Examples
    --------
    Calculate the Gamma distribution with quartiles 3, 6 and 8:

    .. plot::
        :context: close-figs
        :include-source: true

        >>>
        >>> import preliz as pz
        >>> pz.style.use('preliz-doc')
        >>> pz.quartile(pz.Gamma(), 3, 6, 8)

    Calculate the HalfStudentT T distribution with quartiles 2, 9 and 12
    and a value of nu=7:

    .. plot::
        :context: close-figs
        :include-source: true

        >>>
        >>> import preliz as pz
        >>> pz.style.use('preliz-doc')
        >>> pz.quartile(pz.HalfStudentT(nu=7), 2, 9, 12)

    """
    valid_distribution(distribution)

    if plot is None:
        plot = rcParams["plots.show_plot"]

    if plot_kwargs is None:
        plot_kwargs = {}

    if not q1 < q2 < q3:
        raise ValueError("The order of the quartiles should be q1 < q2 < q3")

    quartiles = np.array([q1, q2, q3])

    if distribution is None:
        distribution = Normal()

    if distribution.is_frozen:
        raise ValueError("All parameters are fixed, at least one should be free")

    distribution._check_endpoints(q1, q3)

    # Find which parameters has been fixed
    none_idx, fixed = get_fixed_params(distribution)

    # Heuristic to provide an initial guess for the optimization step
    # We obtain those guesses by first approximating the mean and standard deviation
    # from the quartiles and then use those values for moment matching
    distribution._fit_moments(mean=q2, sigma=(q3 - q1) / 1.35)

    opt = optimize_quartile(distribution, quartiles, none_idx, fixed)

    distribution.opt = opt

    r_error, _ = relative_error(distribution, q1, q3, 0.5)

    if r_error > 0.01:
        computed_masses = distribution.cdf(quartiles).astype(float)
        warnings.warn(
            f"\nThe expected masses are 0.25, 0.5, 0.75\n"
            f"The computed ones are: {computed_masses[0]:.2g}, "
            f"{computed_masses[1]:.2g}, {computed_masses[2]:.2g}",
            stacklevel=2,
        )

    if plot:
        ax = distribution.plot_pdf(**plot_kwargs)
        if plot_kwargs.get("pointinterval"):
            cid = -4
        else:
            cid = -1
        ax.plot(quartiles, [0, 0, 0], "o", color=ax.get_lines()[cid].get_c(), alpha=0.5)
        return distribution, ax

    return distribution



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from scipy.stats import kurtosis, skew

from preliz.distributions import Censored, Normal, Poisson


@pytest.mark.parametrize(
    "dist, lower, upper",
    [
        (Normal(0, 2), -2, 2),
        (Poisson(3.5), 1, 6),
    ],
)
def test_censored(dist, lower, upper):
    cen_dist = Censored(dist, lower, upper)
    cen_dist_inf = Censored(dist, -np.inf, np.inf)

    x_vals = cen_dist.rvs(1000000, random_state=1)
    assert_almost_equal(np.mean(x_vals == lower), dist.cdf(lower), decimal=2)
    if dist.kind == "discrete":
        assert_almost_equal(np.mean(x_vals == upper), 1 - dist.cdf(upper - 1), decimal=2)
    else:
        assert_almost_equal(np.mean(x_vals == upper), 1 - dist.cdf(upper), decimal=2)

    x_inside = x_vals[(x_vals > lower) & (x_vals < upper)]
    assert_almost_equal(dist.logpdf(x_inside), cen_dist.logpdf(x_inside))
    assert_almost_equal(dist.cdf(x_inside), cen_dist.cdf(x_inside))

    assert_almost_equal(dist.cdf(x_inside), cen_dist.cdf(x_inside))

    assert_almost_equal(cen_dist.median(), dist.median())
    assert_almost_equal(x_vals.mean(), cen_dist.mean(), decimal=1)
    assert_almost_equal(x_vals.var(), cen_dist.var(), decimal=1)
    assert_almost_equal(skew(x_vals), cen_dist.skewness(), decimal=0)
    assert_almost_equal(kurtosis(x_vals), cen_dist.kurtosis(), decimal=0)

    actual_mean = dist.mean()
    expected_mean = cen_dist_inf.mean()
    assert_almost_equal(actual_mean, expected_mean, decimal=2)

    actual_var = dist.var()
    expected_var = cen_dist_inf.var()
    assert_almost_equal(actual_var, expected_var, decimal=2)

    actual_entropy = dist.entropy()
    expected_entropy = cen_dist_inf.entropy()
    assert_almost_equal(actual_entropy, expected_entropy, decimal=1)

    c_l, c_u = cen_dist.hdi()
    d_l, d_u = dist.hdi()
    assert c_l >= d_l
    assert c_u <= d_u
    assert_almost_equal(cen_dist_inf.hdi(), dist.hdi())



import numpy as np
from scipy.integrate import quad
from scipy.optimize import bisect
from scipy.special import i0e, i1e
from scipy.stats import circmean

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import find_kappa, optimize_moments


class VonMises(Continuous):
    r"""
    Univariate VonMises distribution.

    The pdf of this distribution is

    .. math::

        f(x \mid \mu, \kappa) =
            \frac{e^{\kappa\cos(x-\mu)}}{2\pi I_0(\kappa)}

    where :math:`I_0` is the modified Bessel function of order 0.

    .. plot::
        :context: close-figs


        from preliz import VonMises, style
        style.use('preliz-doc')
        mus = [0., 0., 0.,  -2.5]
        kappas = [.01, 0.5, 4., 2.]
        for mu, kappa in zip(mus, kappas):
            VonMises(mu, kappa).plot_pdf(support=(-np.pi,np.pi))

    ========  ==========================================
    Support   :math:`x \in [-\pi, \pi]`
    Mean      :math:`\mu`
    Variance  :math:`1-\frac{I_1(\kappa)}{I_0(\kappa)}`
    ========  ==========================================

    Parameters
    ----------
    mu : float
        Mean.
    kappa : float
        Concentration (:math:`\frac{1}{\kappa}` is analogous to :math:`\kappa^2`).
    """

    def __init__(self, mu=None, kappa=None):
        super().__init__()
        self._parametrization(mu, kappa)

    def _parametrization(self, mu=None, kappa=None):
        self.mu = mu
        self.kappa = kappa
        self.param_names = ("mu", "kappa")
        self.params_support = ((-np.pi, np.pi), (eps, np.inf))
        self.support = (-np.pi, np.pi)
        if all_not_none(mu, kappa):
            self._update(mu, kappa)

    def _update(self, mu, kappa):
        self.mu = np.float64(mu)
        self.kappa = np.float64(kappa)
        self.params = (self.mu, self.kappa)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        return nb_cdf(x, self.pdf)

    def ppf(self, q):
        return nb_ppf(q, self.pdf)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.kappa)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.kappa)

    def entropy(self):
        return nb_entropy(self.kappa, self.var())

    def mean(self):
        return self.mu

    def mode(self):
        return self.mu

    def median(self):
        return self.mu

    def var(self):
        return 1 - i1e(self.kappa) / i0e(self.kappa)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 0

    def kurtosis(self):
        return 0

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.vonmises(self.mu, self.kappa, size)

    def _fit_moments(self, mean, sigma):
        params = mean, 1 / sigma**1.8
        optimize_moments(self, mean, sigma, params)

    def _fit_mle(self, sample):
        data = np.mod(sample, 2 * np.pi)
        mu = circmean(data)
        kappa = find_kappa(data, mu)
        mu = np.mod(mu + np.pi, 2 * np.pi) - np.pi
        self._update(mu, kappa)

    def eti(self, mass=None, fmt=".2f"):
        mean = self.mu
        self.mu = 0
        hdi_min, hdi_max = super().eti(mass=mass, fmt=fmt)
        self.mu = mean
        return _warp_interval(hdi_min, hdi_max, self.mu, fmt)

    def hdi(self, mass=None, fmt=".2f"):
        mean = self.mu
        self.mu = 0
        hdi_min, hdi_max = super().hdi(mass=mass, fmt=fmt)
        self.mu = mean
        return _warp_interval(hdi_min, hdi_max, self.mu, fmt)


def nb_cdf(x, pdf):
    if isinstance(x, int | float):
        x = [x]
        scalar_input = True
    else:
        scalar_input = False

    cdf_values = np.array([quad(pdf, -np.pi, xi)[0] if xi <= np.pi else 1 for xi in x])

    return cdf_values[0] if scalar_input else cdf_values


def nb_ppf(q, pdf):
    def root_func(x, q):
        return nb_cdf(x, pdf) - q

    if isinstance(q, int | float):
        q = [q]
        scalar_input = True
    else:
        scalar_input = False

    ppf_values = []
    for q_i in q:
        if q_i < 0:
            val = np.nan
        elif q_i > 1:
            val = np.nan
        elif q_i == 0:
            val = -np.inf
        elif q_i == 1:
            val = np.inf
        else:
            val = bisect(root_func, -np.pi, np.pi, args=(q_i,))

        ppf_values.append(val)

    return ppf_values[0] if scalar_input else np.array(ppf_values)


def nb_entropy(kappa, var):
    return np.log(2 * np.pi * i0e(kappa)) + kappa * var


def nb_logpdf(x, mu, kappa):
    return kappa * (np.cos(x - mu) - 1) - np.log(2 * np.pi) - np.log(i0e(kappa))


def nb_neg_logpdf(x, mu, kappa):
    return -(nb_logpdf(x, mu, kappa)).sum()


def _warp_interval(hdi_min, hdi_max, mu, fmt):
    hdi_min = hdi_min + mu
    hdi_max = hdi_max + mu

    lower_tail = np.arctan2(np.sin(hdi_min), np.cos(hdi_min))
    upper_tail = np.arctan2(np.sin(hdi_max), np.cos(hdi_max))
    if fmt != "none":
        lower_tail = float(f"{lower_tail:{fmt}}")
        upper_tail = float(f"{upper_tail:{fmt}}")
    return (lower_tail, upper_tail)



from math import ceil, floor

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches

try:
    import ipywidgets as widgets
except ImportError:
    pass

from preliz.distributions import all_continuous, all_discrete
from preliz.internal.distribution_helper import get_distributions, process_extra
from preliz.internal.optimization import fit_to_epdf
from preliz.internal.plot_helper import check_inside_notebook, representations


class Roulette:
    """
    Prior elicitation for 1D distribution using the roulette method (See :footcite:t:`Morris2014`).

    Draw 1D distributions using a grid as input.

    Parameters
    ----------
    x_min: Optional[float]
        Minimum value for the domain of the grid and fitted distribution.
    x_max: Optional[float]
        Maximum value for the domain of the grid and fitted distribution.
    nrows: Optional[int]
        Number of rows for the grid. Defaults to 10.
    ncols: Optional[int]
        Number of columns for the grid. Defaults to 11.
    dist_names: list
        List of distribution names to be used in the elicitation.
        Defaults to None. The pre-selected distributions are ["Normal", "BetaScaled",
        "Gamma", "LogNormal", "StudentT"], but almost all 1D PreliZ's distributions
        are available to be selected from the menu with some exceptions like Uniform
        or Cauchy.
    params: Optional[str]
        Extra parameters to be passed to the distributions. The format is a string with the
        PreliZ's distribution name followed by the argument to fix.
        For example: "TruncatedNormal(lower=0), StudentT(nu=8)". If you use the ``params``
        text area, quotation marks are not necessary.
    figsize: Optional[Tuple[int, int]]
        Figure size. If None, it will be defined automatically.

    Returns
    -------
    Roulette object
        The object has many attributes, but the most important are:
        - dist: The fitted distribution.
        - inputs: A tuple with the x values, the empirical pdf, the total
            chips, the x_min, the x_max, the number of rows, and the number of columns.

    """

    def __init__(
        self, x_min=0, x_max=10, nrows=10, ncols=11, dist_names=None, params=None, figsize=None
    ):
        self._x_min = x_min
        self._x_max = x_max
        self._nrows = nrows
        self._ncols = ncols
        self._dist_names = dist_names
        self._figsize = figsize
        self._w_extra = params
        self.dist = None
        self.inputs = None

        check_inside_notebook(need_widget=True)

        self._widgets = self._get_widgets()
        self._output = widgets.Output()

        with self._output:
            if self._figsize is None:
                self._figsize = (8, 6)

            self._fig, self._ax_grid, self._ax_fit = self._create_figure()
            self._coll = self._create_grid()
            self._grid = _Rectangles(self._fig, self._coll, self._nrows, self._ncols, self._ax_grid)

            self._setup_observers()

            self._fig.canvas.mpl_connect("button_release_event", lambda event: self._on_leave_fig())

        controls = widgets.VBox(
            [
                self._widgets["w_x_min"],
                self._widgets["w_x_max"],
                self._widgets["w_nrows"],
                self._widgets["w_ncols"],
                self._widgets["w_extra"],
            ]
        )
        control_distribution = widgets.VBox(
            [
                self._widgets["w_checkbox_cont"],
                self._widgets["w_checkbox_disc"],
                self._widgets["w_checkbox_none"],
            ]
        )
        display(  # noqa: F821
            widgets.HBox(
                [
                    controls,
                    self._widgets["w_repr"],
                    self._widgets["w_distributions"],
                    control_distribution,
                ]
            )
        )

    def _create_figure(self):
        fig, axes = plt.subplots(2, 1, figsize=self._figsize, constrained_layout=True)
        ax_grid = axes[0]
        ax_fit = axes[1]
        ax_fit.set_yticks([])
        fig.canvas.header_visible = False
        fig.canvas.footer_visible = False
        fig.canvas.toolbar_position = "right"
        return fig, ax_grid, ax_fit

    def _create_grid(self):
        xx = np.arange(self._ncols)
        yy = np.arange(self._nrows)

        if self._ncols < 11:
            num = self._ncols
        else:
            num = 11

        self._ax_grid.set(
            xticks=np.linspace(0, self._ncols - 1, num=num) + 0.5,
            xticklabels=[f"{i:.1f}" for i in np.linspace(self._x_min, self._x_max, num=num)],
        )

        coll = np.zeros((self._nrows, self._ncols), dtype=object)
        for idx, xi in enumerate(xx):
            for idy, yi in enumerate(yy):
                sq = patches.Rectangle((xi, yi), 1, 1, fill=True, facecolor="0.8", edgecolor="w")
                self._ax_grid.add_patch(sq)
                coll[idy, idx] = sq

        self._ax_grid.set_yticks([])
        self._ax_grid.relim()
        self._ax_grid.autoscale_view()
        return coll

    def _on_leave_fig(self):
        extra_pros = process_extra(self._widgets["w_extra"].value)

        x_vals, epdf, mean, std, filled_columns = self._weights_to_pdf()

        fitted_dist = None
        if filled_columns > 1:
            selected_distributions = get_distributions(self._widgets["w_distributions"].value)

            if selected_distributions:
                self._reset_dist_panel(yticks=False)
                fitted_dist = fit_to_epdf(
                    selected_distributions,
                    x_vals,
                    epdf,
                    mean,
                    std,
                    self._x_min,
                    self._x_max,
                    extra_pros,
                )

                if fitted_dist is None:
                    self._ax_fit.set_title("domain error")
                else:
                    representations(fitted_dist, self._widgets["w_repr"].value, self._ax_fit)
        else:
            self._reset_dist_panel(yticks=True)
        self._fig.canvas.draw()

        self.inputs = (
            x_vals,
            epdf,
            sum(self._grid._weights.values()),
            self._x_min,
            self._x_max,
            self._nrows,
            self._ncols,
        )
        self.dist = fitted_dist

    def _weights_to_pdf(self):
        step = (self._x_max - self._x_min) / (self._ncols - 1)
        x_vals = [(k + 0.5) * step + self._x_min for k, v in self._grid._weights.items() if v != 0]
        total = sum(self._grid._weights.values())
        epdf = [v / total for v in self._grid._weights.values() if v != 0]

        mean = sum(prob * value for value, prob in zip(x_vals, epdf))
        std = (sum(prob * (value - mean) ** 2 for value, prob in zip(x_vals, epdf))) ** 0.5

        return x_vals, epdf, mean, std, len(x_vals)

    def _update_grid(self):
        self._ax_grid.cla()
        self._coll = self._create_grid()
        self._grid._coll = self._coll
        self._grid._ncols = self._ncols
        self._grid._nrows = self._nrows
        self._grid._weights = {k: 0 for k in range(0, self._ncols)}
        self._reset_dist_panel(yticks=True)
        self._ax_grid.set_yticks([])
        self._ax_grid.relim()
        self._ax_grid.autoscale_view()
        self._fig.canvas.draw()

    def _reset_dist_panel(self, yticks):
        self._ax_fit.cla()
        if yticks:
            self._ax_fit.set_yticks([])
        self._ax_fit.set_xlim(self._x_min, self._x_max)
        self._ax_fit.relim()
        self._ax_fit.autoscale_view()

    def _handle_checkbox_widget(self):
        if self._widgets["w_checkbox_none"].value:
            self._widgets["w_checkbox_disc"].value = False
            self._widgets["w_checkbox_cont"].value = False
            return []
        all_cls = []
        if self._widgets["w_checkbox_cont"].value:
            all_cls += list(
                cls.__name__
                for cls in all_continuous
                if cls.__name__ in self._widgets["w_distributions"].options
            )
        if self._widgets["w_checkbox_disc"].value:
            all_cls += list(
                cls.__name__
                for cls in all_discrete
                if cls.__name__ in self._widgets["w_distributions"].options
            )
        return all_cls

    def _get_widgets(self):
        width_entry_text = widgets.Layout(width="150px")
        width_repr_text = widgets.Layout(width="250px")
        width_distribution_text = widgets.Layout(width="150px", height="125px")

        w_x_min = widgets.FloatText(
            value=self._x_min,
            step=1,
            description="x_min:",
            disabled=False,
            layout=width_entry_text,
        )

        w_x_max = widgets.FloatText(
            value=self._x_max,
            step=1,
            description="x_max:",
            disabled=False,
            layout=width_entry_text,
        )

        w_nrows = widgets.BoundedIntText(
            value=self._nrows,
            min=2,
            step=1,
            description="n_rows:",
            disabled=False,
            layout=width_entry_text,
        )

        w_ncols = widgets.BoundedIntText(
            value=self._ncols,
            min=2,
            step=1,
            description="n_cols:",
            disabled=False,
            layout=width_entry_text,
        )

        w_extra = widgets.Textarea(
            value=self._w_extra,
            placeholder="Pass extra parameters",
            description="params:",
            disabled=False,
            layout=width_repr_text,
        )

        w_repr = widgets.RadioButtons(
            options=["pdf", "cdf", "ppf"],
            value="pdf",
            description="",
            disabled=False,
            layout=width_entry_text,
        )

        if self._dist_names is None:
            default_dist = ["Normal", "BetaScaled", "Gamma", "LogNormal", "StudentT"]
            dist_names = [
                "AsymmetricLaplace",
                "BetaScaled",
                "ChiSquared",
                "ExGaussian",
                "Exponential",
                "Gamma",
                "Gumbel",
                "HalfNormal",
                "HalfStudentT",
                "InverseGamma",
                "Laplace",
                "LogNormal",
                "Logistic",
                # "LogitNormal", # fails if we add chips at x_value= 1
                "Moyal",
                "Normal",
                "Pareto",
                "Rice",
                "SkewNormal",
                "StudentT",
                "Triangular",
                "VonMises",
                "Wald",
                "Weibull",
                "BetaBinomial",
                "DiscreteWeibull",
                "Geometric",
                "NegativeBinomial",
                "Poisson",
            ]
        else:
            default_dist = self._dist_names
            dist_names = self._dist_names

        w_distributions = widgets.SelectMultiple(
            options=dist_names,
            value=default_dist,
            description="",
            disabled=False,
            layout=width_distribution_text,
        )

        w_checkbox_cont = widgets.Checkbox(
            value=False, description="Continuous", disabled=False, indent=False
        )
        w_checkbox_disc = widgets.Checkbox(
            value=False, description="Discrete", disabled=False, indent=False
        )
        w_checkbox_none = widgets.Checkbox(
            value=False, description="None", disabled=False, indent=False
        )

        return {
            "w_x_min": w_x_min,
            "w_x_max": w_x_max,
            "w_ncols": w_ncols,
            "w_nrows": w_nrows,
            "w_extra": w_extra,
            "w_repr": w_repr,
            "w_distributions": w_distributions,
            "w_checkbox_cont": w_checkbox_cont,
            "w_checkbox_disc": w_checkbox_disc,
            "w_checkbox_none": w_checkbox_none,
        }

    def _setup_observers(self):
        self._widgets["w_checkbox_none"].observe(self._handle_checkbox_change)
        self._widgets["w_checkbox_cont"].observe(self._handle_checkbox_change)
        self._widgets["w_checkbox_disc"].observe(self._handle_checkbox_change)

        def _update_grid_(_):
            self._x_min = self._widgets["w_x_min"].value
            self._x_max = self._widgets["w_x_max"].value
            self._nrows = self._widgets["w_nrows"].value
            self._ncols = self._widgets["w_ncols"].value
            self._update_grid()

        self._widgets["w_x_min"].observe(_update_grid_)
        self._widgets["w_x_max"].observe(_update_grid_)
        self._widgets["w_nrows"].observe(_update_grid_)
        self._widgets["w_ncols"].observe(_update_grid_)
        self._widgets["w_x_min"].observe(self._on_value_change, names="value")

        def _on_leave_fig_(_):
            self._on_leave_fig()

        self._widgets["w_repr"].observe(_on_leave_fig_)
        self._widgets["w_distributions"].observe(_on_leave_fig_)
        self._widgets["w_extra"].observe(_on_leave_fig_)

    def _handle_checkbox_change(self, _):
        dist_names = self._handle_checkbox_widget()
        self._widgets["w_distributions"].value = dist_names

    def _on_value_change(self, change):
        new_a = change["new"]
        if new_a == self._widgets["w_x_max"].value:
            self._widgets["w_x_max"].value = new_a + 1


class _Rectangles:
    def __init__(self, fig, coll, nrows, ncols, ax):
        self._fig = fig
        self._coll = coll
        self._nrows = nrows
        self._ncols = ncols
        self._ax = ax
        self._weights = {k: 0 for k in range(0, ncols)}
        fig.canvas.mpl_connect("button_press_event", self)

    def __call__(self, event):
        if event.inaxes == self._ax:
            x = event.xdata
            y = event.ydata
            idx = floor(x)
            idy = ceil(y)

            if 0 <= idx < self._ncols and 0 <= idy <= self._nrows:
                if self._weights[idx] >= idy:
                    idy -= 1
                    for row in range(self._nrows):
                        self._coll[row, idx].set_facecolor("0.8")
                self._weights[idx] = idy
                for row in range(idy):
                    self._coll[row, idx].set_facecolor("C1")
                self._fig.canvas.draw()



import warnings

from preliz.distributions.normal import Normal
from preliz.internal.distribution_helper import valid_distribution
from preliz.internal.optimization import get_fixed_params, optimize_max_ent, relative_error
from preliz.internal.rcparams import rcParams


def maxent(
    distribution=None,
    lower=-1,
    upper=1,
    mass=None,
    mode=None,
    fixed_stat=None,
    plot=None,
    plot_kwargs=None,
    ax=None,
):
    """
    Find the maximum entropy distribution that satisfies the constraints.

    Find the maximum entropy distribution with `mass` in the interval
    defined by the `lower` and `upper` end-points.

    Parameters
    ----------
    name : PreliZ distribution
        Instance of a PreliZ distribution. Notice that the distribution will be
        updated inplace.
    lower : float
        Lower end-point
    upper: float
        Upper end-point
    mass: float
        Probability mass between ``lower`` and ``upper`` bounds. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.
    mode: float
        Mode of the distribution. Pass a value to fix the mode of the distribution.
    fixed_stat: tuple
        Summary statistic to fix. The first element should be a name and the second a
        numerical value. Valid names are: "mean", "mode", "median", "variance", "std",
        "skewness", "kurtosis". Defaults to None.
    plot : bool
        Whether to plot the distribution, and lower and upper bounds. Defaults to None,
        which results in the value of rcParams["plots.show_plot"] being used.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()`` of ``distribution``.
    ax : matplotlib axes

    Returns
    -------
    dict: dict with the parameters of the distribution
    axes: matplotlib axes (only if `plot=True`)

    Notes
    -----
    After calling this function the attribute `opt` of the distribution will be updated with the
    OptimizeResult object from the optimization step.

    See Also
    --------
    quartile : Find the distribution with the specified quartiles.

    Examples
    --------
    Calculate the maxent Gamma distribution with 90 % of the mass between 1 and 8:

    .. plot::
        :context: close-figs
        :include-source: true

        >>>
        >>> import preliz as pz
        >>> pz.style.use('preliz-doc')
        >>> pz.maxent(pz.Gamma(), 1, 8, 0.9)

    Calculate the maxent HalfStudentT T distribution with 90 % of the mass between 0 and 12
    and a value of nu=4:

    .. plot::
        :context: close-figs
        :include-source: true

        >>>
        >>> import preliz as pz
        >>> pz.style.use('preliz-doc')
        >>> pz.maxent(pz.HalfStudentT(nu=4), 0, 12, 0.9)
    """
    valid_distribution(distribution)

    if mode is not None:
        fixed_stat = ("mode", mode)
        warnings.warn(
            "The parameter `mode` is deprecated and will be removed in a future release. "
            "Use `fixed_stat=('mode', mode)` instead.",
            FutureWarning,
        )

    if mass is None:
        mass = rcParams["stats.ci_prob"]

    if plot is None:
        plot = rcParams["plots.show_plot"]

    if fixed_stat is None:
        fixed_stat = ()
    else:
        valid_stats = ["mean", "mode", "median", "variance", "std", "skewness", "kurtosis"]
        if fixed_stat[0] not in valid_stats:
            raise ValueError("fixed_stat should be one of the following: " + ", ".join(valid_stats))

    if not 0 < mass <= 1:
        raise ValueError("mass should be larger than 0 and smaller or equal to 1")

    if upper <= lower:
        raise ValueError("upper should be larger than lower")

    if plot_kwargs is None:
        plot_kwargs = {}

    if distribution is None:
        distribution = Normal()

    if distribution.is_frozen:
        raise ValueError("All parameters are fixed, at least one should be free")

    distribution._check_endpoints(lower, upper)

    if distribution.kind == "discrete":
        if not end_points_ints(lower, upper):
            warnings.warn(
                f"\n{distribution.__class__.__name__} distribution is discrete, "
                "but the provided bounds are not integers"
            )

    # Find which parameters has been fixed
    none_idx, fixed_params = get_fixed_params(distribution)

    # Heuristic to provide an initial guess for the optimization step
    # We obtain those guesses by first approximating the mean and standard deviation
    # from intervals and mass and then use those values for moment matching
    if distribution.__class__.__name__ == "Uniform":
        distribution._fit_moments(mean=(lower + upper) / 2, sigma=((upper - lower) / 3.4) / mass)
    else:
        distribution._fit_moments(mean=(lower + upper) / 2, sigma=((upper - lower) / 4) / mass)

    if "mode" in fixed_stat:
        try:
            distribution.mode()
        except NotImplementedError as exc:
            raise ValueError(
                f"{distribution.__class__.__name__} does not have a mode method"
            ) from exc

    opt = optimize_max_ent(distribution, lower, upper, mass, none_idx, fixed_params, fixed_stat)
    distribution.opt = opt

    r_error, computed_mass = relative_error(distribution, lower, upper, mass)

    if r_error > 0.01:
        warnings.warn(
            f"\nThe requested mass is {mass:.3g},\n" f"but the computed one is {computed_mass:.3g}",
            stacklevel=2,
        )

    if plot:
        ax = distribution.plot_pdf(**plot_kwargs)
        if plot_kwargs.get("pointinterval"):
            cid = -4
        else:
            cid = -1
        ax.plot([lower, upper], [0, 0], "o", color=ax.get_lines()[cid].get_c(), alpha=0.5)
        return distribution, ax

    return distribution


def end_points_ints(lower, upper):
    return is_integer_num(lower) and is_integer_num(upper)


def is_integer_num(obj):
    if isinstance(obj, int):
        return True
    if isinstance(obj, float):
        return obj.is_integer()
    return False



import numpy as np
from scipy.special import chndtr, chndtrix, i0, i0e, i1

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml, optimize_moments_rice
from preliz.internal.special import cdf_bounds, ppf_bounds_cont


class Rice(Continuous):
    r"""
    Rice distribution.

    The pdf of this distribution is

    .. math::

        f(x\mid \nu ,\sigma )=
            {\frac  {x}{\sigma ^{2}}}\exp
            \left({\frac  {-(x^{2}+\nu ^{2})}
            {2\sigma ^{2}}}\right)I_{0}\left({\frac  {x\nu }{\sigma ^{2}}}\right)

    .. plot::
        :context: close-figs


        from preliz import Rice, style
        style.use('preliz-doc')
        nus = [0., 0., 4.]
        sigmas = [1., 2., 2.]
        for nu, sigma in  zip(nus, sigmas):
            Rice(nu, sigma).plot_pdf(support=(0,10))

    ========  ==============================================================
    Support   :math:`x \in (0, \infty)`
    Mean      :math:`\sigma \sqrt{\pi /2} L_{1/2}(-\nu^2 / 2\sigma^2)`
    Variance  :math:`2\sigma^2 + \nu^2 - \frac{\pi \sigma^2}{2}`
              :math:`L_{1/2}^2\left(\frac{-\nu^2}{2\sigma^2}\right)`
    ========  ==============================================================

    Rice distribution has 2 alternative parameterizations. In terms of nu and sigma
    or b and sigma.

    The link between the two parametrizations is given by

    .. math::

       b = \dfrac{\nu}{\sigma}

    Parameters
    ----------
    nu : float
        Noncentrality parameter.
    sigma : float
        Scale parameter.
    b : float
        Shape parameter.
    """

    def __init__(self, nu=None, sigma=None, b=None):
        super().__init__()
        self.name = "rice"
        self.support = (0, np.inf)
        self._parametrization(nu, sigma, b)

    def _parametrization(self, nu=None, sigma=None, b=None):
        if all_not_none(nu, b):
            raise ValueError(
                "Incompatible parametrization. Either use nu and sigma or b and sigma."
            )

        self.param_names = ("nu", "sigma")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if b is not None:
            self.b = b
            self.sigma = sigma
            self.param_names = ("b", "sigma")
            if all_not_none(b, sigma):
                nu = self._from_b(b, sigma)

        self.nu = nu
        self.sigma = sigma
        if all_not_none(self.nu, self.sigma):
            self._update(self.nu, self.sigma)

    def _from_b(self, b, sigma):
        nu = b * sigma
        return nu

    def _to_b(self, nu, sigma):
        b = nu / sigma
        return b

    def _update(self, nu, sigma):
        self.nu = np.float64(nu)
        self.sigma = np.float64(sigma)
        self.b = self._to_b(self.nu, self.sigma)

        if self.param_names[0] == "nu":
            self.params = (self.nu, self.sigma)
        elif self.param_names[0] == "b":
            self.params = (self.b, self.sigma)

        self.is_frozen = True

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.nu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.nu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.nu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.nu, self.sigma)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return self.sigma * np.sqrt(np.pi / 2) * _l_half(-(self.nu**2) / (2 * self.sigma**2))

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (
            2 * self.sigma**2
            + self.nu**2
            - np.pi / 2 * self.sigma**2 * _l_half(-(self.nu**2) / (2 * self.sigma**2)) ** 2
        )

    def std(self):
        return self.var() ** 2

    def skewness(self):
        return NotImplemented

    def kurtosis(self):
        return NotImplemented

    def rvs(self, size=1, random_state=None):
        random_state = np.random.default_rng(random_state)
        t_v = (self.nu / self.sigma) / np.sqrt(2) + random_state.standard_normal(size=(2, size))
        return np.sqrt((t_v * t_v).sum(axis=0)) * self.sigma

    def _fit_moments(self, mean, sigma):
        nu, sigma = optimize_moments_rice(mean, sigma)
        self._update(nu, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


def nb_cdf(x, nu, sigma):
    return cdf_bounds(chndtr((x / sigma) ** 2, 2, (nu / sigma) ** 2), x, 0, np.inf)


def nb_ppf(q, nu, sigma):
    return ppf_bounds_cont(np.sqrt(chndtrix(q, 2, (nu / sigma) ** 2)) * sigma, q, 0, np.inf)


def nb_logpdf(x, nu, sigma):
    b = nu / sigma
    x = x / sigma
    return np.where(
        x < 0, -np.inf, np.log(x * np.exp((-(x - b) * (x - b)) / 2) * i0e(x * b) / sigma)
    )


def nb_neg_logpdf(x, nu, sigma):
    return -(nb_logpdf(x, nu, sigma)).sum()


def _l_half(x):
    return np.exp(x / 2) * ((1 - x) * i0(-x / 2) - x * i1(-x / 2))



import numba as nb
import numpy as np

from preliz.distributions.distributions import Discrete
from preliz.internal.distribution_helper import eps
from preliz.internal.special import cdf_bounds, mean_sample, ppf_bounds_disc, xlog1py, xlogx


class Geometric(Discrete):
    R"""
    Geometric distribution.

    The probability that the first success in a sequence of Bernoulli trials
    occurs on the x'th trial.
    The pmf of this distribution is

    .. math::
        f(x \mid p) = p(1-p)^{x-1}

    .. plot::
        :context: close-figs


        from preliz import Geometric, style
        style.use('preliz-doc')
        for p in [0.1, 0.25, 0.75]:
            Geometric(p).plot_pdf(support=(1,10))

    ========  =============================
    Support   :math:`x \in \mathbb{N}_{>0}`
    Mean      :math:`\dfrac{1}{p}`
    Variance  :math:`\dfrac{1 - p}{p^2}`
    ========  =============================

    Parameters
    ----------
    p : float
        Probability of success on an individual trial (0 < p <= 1).
    """

    def __init__(self, p=None):
        super().__init__()
        self.support = (1, np.inf)
        self._parametrization(p)

    def _parametrization(self, p=None):
        self.p = p
        self.param_names = "p"
        self.params_support = ((eps, 1),)
        if self.p is not None:
            self._update(self.p)

    def _update(self, p):
        self.p = np.float64(p)
        self.params = (self.p,)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.p))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.p, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.p, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.p)

    def entropy(self):
        return nb_entropy(self.p)

    def mean(self):
        return 1 / self.p

    def mode(self):
        return np.ones_like(self.p)

    def median(self):
        return np.ceil(-1 / np.log(1 - self.p))

    def var(self):
        return (1 - self.p) / self.p**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (2 - self.p) / (1 - self.p) ** 0.5

    def kurtosis(self):
        return 6 + (self.p**2) / (1 - self.p)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.geometric(self.p, size=size)

    def _fit_moments(self, mean, sigma):
        p = 1 / mean
        self._update(p)

    def _fit_mle(self, sample):
        p = 1 / mean_sample(sample)
        self._update(p)


@nb.njit(cache=True)
def nb_cdf(x, p, lower, upper):
    x = np.floor(x)
    prob = 1 - (1 - p) ** x
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, p, lower, upper):
    x_vals = np.ceil(np.log(1 - q) / np.log(1 - p))
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(p):
    return (xlog1py((-1 + p), -p) - xlogx(p)) / p


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, p):
    if x < 1:
        return -np.inf
    else:
        return xlog1py(x - 1, -p) + np.log(p)


@nb.njit(cache=True)
def nb_neg_logpdf(x, p):
    return -(nb_logpdf(x, p)).sum()



import numba as nb
import numpy as np

from preliz.distributions.distributions import Discrete
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml
from preliz.internal.special import expit, logit, xlogx


class Bernoulli(Discrete):
    R"""Bernoulli distribution.

    The Bernoulli distribution describes the probability of successes (x=1) and failures (x=0).
    The pmf of this distribution is

    .. math::
        f(x \mid p) = p^{x} (1-p)^{1-x}

    .. plot::
        :context: close-figs


        from preliz import Bernoulli, style
        style.use('preliz-doc')
        for p in [0, 0.5, 0.8]:
            Bernoulli(p).plot_pdf()

    ========  ======================
    Support   :math:`x \in \{0, 1\}`
    Mean      :math:`p`
    Variance  :math:`p (1 - p)`
    ========  ======================

    The Bernoulli distribution has 2 alternative parametrizations. In terms of p or logit_p.

    The link between the 2 alternatives is given by

    .. math::

        logit(p) = ln(\frac{p}{1-p})

    Parameters
    ----------
    p : float
        Probability of success (0 < p < 1).
    logit_p : float
        Alternative log odds for the probability of success.
    """

    def __init__(self, p=None, logit_p=None):
        super().__init__()
        self.support = (0, 1)
        self._parametrization(p, logit_p)

    def _parametrization(self, p=None, logit_p=None):
        if all_not_none(p, logit_p):
            raise ValueError("Incompatible parametrization. Either use p or logit_p.")

        self.param_names = "p"
        self.params_support = ((eps, 1),)

        if logit_p is not None:
            p = self._from_logit_p(logit_p)
            self.param_names = ("logit_p",)

        self.p = p
        self.logit_p = logit_p
        if self.p is not None:
            self._update(self.p)

    def _from_logit_p(self, logit_p):
        return expit(logit_p)

    def _to_logit_p(self, p):
        return logit(p)

    def _update(self, p):
        self.p = np.float64(p)
        self._q = 1 - self.p
        self.logit_p = self._to_logit_p(p)

        if self.param_names[0] == "p":
            self.params = (self.p,)
        elif self.param_names[0] == "logit_p":
            self.params = (self.logit_p,)

        self.is_frozen = True

    def _fit_moments(self, mean, sigma):
        self._update(mean)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)

    def pdf(self, x):
        x = np.asarray(x)
        return nb_pdf(x, self.p)

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.p)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.p)

    def logpdf(self, x):
        x = np.asarray(x)
        return nb_logpdf(x, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.p)

    def entropy(self):
        return nb_entropy(self.p)

    def mean(self):
        return self.p

    def mode(self):
        return self.median()

    def median(self):
        return np.where(self.p <= 0.5, 0, 1)

    def var(self):
        return self.p * self._q

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (self._q - self.p) / self.std()

    def kurtosis(self):
        return (1 - 6 * self.p * self._q) / (self.p * self._q)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.binomial(1, self.p, size=size)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, p):
    if x < 0:
        return 0
    elif x < 1:
        return 1 - p
    else:
        return 1


@nb.vectorize(nopython=True, cache=True)
def nb_ppf(q, p):
    if q < 0:
        return np.nan
    elif q > 1:
        return np.nan
    elif q == 0:
        return -1
    elif q < 1 - p:
        return 0
    else:
        return 1


@nb.vectorize(nopython=True, cache=True)
def nb_pdf(x, p):
    if x == 1:
        return p
    elif x == 0:
        return 1 - p
    else:
        return 0.0


@nb.njit(cache=True)
def nb_entropy(p):
    q = 1 - p
    return -xlogx(q) - xlogx(p)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, p):
    if x == 1:
        return np.log(p)
    elif x == 0:
        return np.log(1 - p)
    else:
        return -np.inf


@nb.njit(cache=True)
def nb_neg_logpdf(x, p):
    return -(nb_logpdf(x, p)).sum()



import os
import re


def citations(methods=None, filepath=None, format_type="bibtex"):
    """
    Generate references associated with the methods.

    Parameters
    ----------
    methods : str or list, optional
        Specifies the PreliZ methods or classes for which to display references.
        Use `"all"` to show references for all available methods/classes.
        Defaults to `None`, which displays PreliZ references.

    filepath : str or a file-like object
        Specifies the location to save the file.
        If ``None``, the result is returned as a string.

    format_type : str
       Specifies in which type the references will be displayed.
       Currently, only "bibtex" is supported.
       Defaults to ``bibtex``.

    """
    ref_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "reference.bib")
    with open(ref_path, encoding="utf-8") as fr:
        content = fr.read()
    if methods == "all":
        keys = set(re.findall(r"@.*?\{(.*?),", content))
    elif methods is None:
        keys = {"PreliZ", "Icazatti2023"}
    else:
        keys = set()
        for method in methods:
            matches = set(re.findall(r":(?:cite|footcite):[tp]:`(.*?)`", method.__doc__))
            keys.update(matches)
    if format_type == "bibtex":
        cite_text = _citation_bibtex(content, keys)
        if filepath:
            with open(filepath, "w") as fw:
                fw.write(cite_text)
        else:
            return cite_text
    else:
        raise ValueError("Invalid value for format_type. Use 'bibtex'.")


def _citation_bibtex(content, keys):
    """Extract and return references in BibTeX format."""
    extracted_refs = []
    for key in keys:
        match = re.search(rf"(@\w+\{{\s*{key}\s*,.*?\n\}})", content, re.DOTALL)
        extracted_refs.append(match.group(1))
    return "\n".join(extracted_refs)



import numba as nb
import numpy as np

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps


class Laplace(Continuous):
    r"""
    Laplace distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu, b) =
           \frac{1}{2b} \exp \left\{ - \frac{|x - \mu|}{b} \right\}

    .. plot::
        :context: close-figs


        from preliz import Laplace, style
        style.use('preliz-doc')
        mus = [0., 0., 0., -5.]
        bs = [1., 2., 4., 4.]
        for mu, b in zip(mus, bs):
            Laplace(mu, b).plot_pdf(support=(-10,10))

    ========  ========================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu`
    Variance  :math:`2 b^2`
    ========  ========================

    Parameters
    ----------
    mu : float
        Location parameter.
    b : float
        Scale parameter (b > 0).
    """

    def __init__(self, mu=None, b=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, b)

    def _parametrization(self, mu=None, b=None):
        self.mu = mu
        self.b = b
        self.params = (self.mu, self.b)
        self.param_names = ("mu", "b")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        if all_not_none(mu, b):
            self._update(mu, b)

    def _update(self, mu, b):
        self.mu = np.float64(mu)
        self.b = np.float64(b)
        self.params = (self.mu, self.b)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.b))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.b)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.b)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.b)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.b)

    def entropy(self):
        return nb_entropy(self.b)

    def median(self):
        return self.mu

    def mean(self):
        return self.mu

    def mode(self):
        return self.mu

    def std(self):
        return self.var() ** 0.5

    def var(self):
        return 2 * self.b**2

    def skewness(self):
        return 0.0

    def kurtosis(self):
        return 3.0

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.laplace(self.mu, self.b, size)

    def _fit_moments(self, mean, sigma):
        b = (sigma / 2) * (2**0.5)
        self._update(mean, b)

    def _fit_mle(self, sample):
        mu, b = nb_fit_mle(sample)
        self._update(mu, b)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, mu, b):
    x = (x - mu) / b
    if x > 0:
        return 1.0 - 0.5 * np.exp(-x)
    return 0.5 * np.exp(x)


@nb.vectorize(nopython=True, cache=True)
def nb_ppf(q, mu, b):
    if q > 0.5:
        q = -np.log(2 * (1 - q))
    else:
        q = np.log(2 * q)
    return q * b + mu


@nb.njit(cache=True)
def nb_logpdf(x, mu, b):
    x = (x - mu) / b
    return np.log(0.5) - np.abs(x) - np.log(b)


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, b):
    return (-nb_logpdf(x, mu, b)).sum()


@nb.njit(cache=True)
def nb_entropy(b):
    return np.log(2) + 1 + np.log(b)


@nb.njit(cache=True)
def nb_fit_mle(sample):
    median = np.median(sample)
    scale = np.sum(np.abs(sample - median)) / len(sample)
    return median, scale



import inspect
import sys
import traceback

try:
    from IPython import get_ipython
    from ipywidgets import Checkbox, FloatSlider, FloatText, IntSlider, IntText, ToggleButton
except ImportError:
    pass

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import _pylab_helpers, get_backend
from matplotlib.ticker import MaxNLocator

from preliz.internal.narviz import hdi, kde
from preliz.internal.rcparams import rcParams


def plot_pointinterval(distribution, interval=None, levels=None, rotated=False, ax=None):
    """
    Plot median as a dot and intervals as lines.

    By defaults the intervals are HDI with 0.5 and the value in rcParams["stats.ci_prob"].

    Parameters
    ----------
    distribution : preliz distribution or array
    interval : str
        Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
    levels : list
        Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
        For quantiles the number of elements should be 5, 3, 1 or 0
        (in this last case nothing will be plotted).
    rotated : bool
        Whether to do the plot along the x-axis (default) or on the y-axis
    ax : matplotlib axis
    """
    if isinstance(distribution, np.ndarray | list | tuple):
        dist_type = "sample"
    else:
        dist_type = "preliz"

    if interval is None:
        interval = rcParams["stats.ci_kind"]

    if interval == "quantiles":
        if levels is None:
            levels = [0.05, 0.25, 0.5, 0.75, 0.95]
        elif len(levels) not in (5, 3, 1, 0):
            raise ValueError("levels should have 5, 3, 1 or 0 elements")

        if dist_type == "sample":
            q_s = np.quantile(distribution, levels).tolist()
        else:
            q_s = distribution.ppf(levels).tolist()

    else:
        if levels is None:
            levels = [0.5, rcParams["stats.ci_prob"]]

        elif len(levels) not in (2, 1):
            raise ValueError("levels should have 2 or 1 elements")

        if dist_type == "sample":
            if interval == "hdi":
                func = hdi
            if interval == "eti":
                func = eti

            q_tmp = np.concatenate([func(distribution, m) for m in levels])
            median = np.median(distribution)
        else:
            if interval == "hdi":
                func = distribution.hdi
            if interval == "eti":
                func = distribution.eti

            q_tmp = np.concatenate([func(mass=m, fmt="none") for m in levels])
            median = distribution.median()

        q_s = []
        if len(levels) == 2:
            q_s.extend((q_tmp[2], q_tmp[0], median, q_tmp[1], q_tmp[3]))
        elif len(levels) == 1:
            q_s.extend((q_tmp[0], median, q_tmp[1]))

    q_s_size = len(q_s)

    if q_s_size == 5:
        _plot_sub_iterval(q_s, lw=1.5, rotated=rotated, ax=ax)
    if q_s_size > 2:
        _plot_sub_iterval(q_s, lw=4, rotated=rotated, ax=ax)
    if q_s_size > 0:
        x, y = q_s[0], 0
        if rotated:
            x, y = y, x
        ax.plot(x, y, "wo", mec="k")


def _plot_sub_iterval(q_s, lw, rotated, ax):
    lower, upper = q_s.pop(0), q_s.pop(-1)
    if lower < upper:
        x, y = (lower, upper), [0, 0]
        if rotated:
            x, y = y, x
        ax.plot(x, y, "k", solid_capstyle="butt", lw=lw)
    else:
        x0, y0 = (lower, np.pi), [0, 0]
        x1, y1 = (-np.pi, upper), [0, 0]
        if rotated:
            x0, y0 = y0, x0
            x1, y1 = y1, x1
        ax.plot(x0, y0, "k", solid_capstyle="butt", lw=lw)
        ax.plot(x1, y1, "k", solid_capstyle="butt", lw=lw)


def eti(distribution, mass):
    lower = (1 - mass) / 2
    return np.quantile(distribution, (lower, mass + lower))


def plot_pdfpmf(
    dist, moments, pointinterval, interval, levels, support, legend, color, alpha, figsize, ax
):
    ax = get_ax(ax, figsize)
    label = set_label(dist, legend, moments, ax)

    x = dist.xvals(support)
    if dist.kind == "continuous":
        if dist.__class__.__name__ == "Censored":
            lower, upper = dist.support
            disc_vals = [None, None]
            if np.min(x) <= lower:
                disc_vals[0] = dist.dist.pdf(lower) * 1.5
                disc_vals[1] = dist.dist.pdf(upper) * 1.5

            x = x[x != dist.support[0]]
            x = x[x != dist.support[1]]

        density = dist.pdf(x)
        ax.axhline(0, color="0.8", ls="--", zorder=0)
        p = ax.plot(x, density, label=label, color=color, alpha=alpha)
        ax.set_yticks([])

        if dist.__class__.__name__ == "Censored":
            if disc_vals[0] is not None:
                ax.vlines(lower, 0, disc_vals[0], ls="--", color=p[0].get_color(), alpha=alpha)
            if disc_vals[1] is not None:
                ax.vlines(upper, 0, disc_vals[1], ls="--", color=p[0].get_color(), alpha=alpha)

    else:
        mass = dist.pdf(x)

        if dist.__class__.__name__ in ["Categorical", "Bernoulli"]:
            p = ax.plot(x, mass, "o", label=label, color=color, alpha=alpha)
            ax.vlines(x, 0, mass, ls="dotted", color=p[0].get_color(), alpha=alpha)
        else:
            x_c = np.linspace(x[0], x[-1], 1000)
            # we compute pmf at non-integer values to get a continuous curve
            mass_c = np.clip(dist.pdf(x_c), np.min(mass), np.max(mass))

            p = ax.plot(x_c, mass_c, ls="dotted", color=color, alpha=alpha)
            ax.plot(x, mass, "o", label=label, color=p[0].get_color(), alpha=alpha)

        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.axhline(0, color="0.8", ls="--", zorder=0)

    if pointinterval:
        plot_pointinterval(dist, interval, levels, ax=ax)

    side_legend(legend, ax)

    return ax


def plot_cdf(
    dist, moments, pointinterval, interval, levels, support, legend, color, alpha, figsize, ax
):
    ax = get_ax(ax, figsize)
    label = set_label(dist, legend, moments, ax)

    ax.set_ylim(-0.05, 1.05)
    x = dist.xvals(support)

    if dist.kind == "discrete":
        lower = x[0]
        upper = x[-1]
        x = np.insert(x, [0, len(x)], (lower - 1, upper + 1))
        cdf = dist.cdf(x)
        ax.set_xlim(lower - 0.1, upper + 0.1)
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.step(x, cdf, where="post", label=label, color=color, alpha=alpha)
    else:
        cdf = dist.cdf(x)
        ax.plot(x, cdf, label=label, color=color, alpha=alpha)

    if pointinterval:
        plot_pointinterval(dist, interval, levels, ax=ax)

    side_legend(legend, ax)

    return ax


def plot_ppf(dist, moments, pointinterval, interval, levels, legend, color, alpha, figsize, ax):
    ax = get_ax(ax, figsize)
    label = set_label(dist, legend, moments, ax)

    x = np.linspace(0, 1, 1000)
    ax.plot(x, dist.ppf(x), label=label, color=color, alpha=alpha)
    if dist.kind == "discrete":
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    if pointinterval:
        plot_pointinterval(dist, interval, levels, rotated=True, ax=ax)

    side_legend(legend, ax)

    return ax


def get_ax(ax, figsize):
    if ax is None:
        fig_manager = _pylab_helpers.Gcf.get_active()
        if fig_manager is not None:
            ax = fig_manager.canvas.figure.gca()
        else:
            _, ax = plt.subplots(figsize=figsize)
    return ax


def side_legend(legend, ax):
    if isinstance(legend, str) and legend != "title":
        bbox = ax.get_position()
        ax.set_position([bbox.x0, bbox.y0, bbox.width, bbox.height])
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))


def repr_to_matplotlib(distribution):
    string = repr(distribution)
    string = string.replace("\x1b[1m", r"$\bf{")
    string = string.replace("\x1b[0m", "}$")
    return string


def get_moments(dist, moments):
    names = {
        "m": "μ",
        "d": "σ",
        "s": "γ",
        "v": "σ²",
        "k": "κ",
    }
    str_m = []
    seen = []
    values = dist.moments(moments)
    for moment, value in zip(moments, values):
        if moment not in seen:
            str_m.append(f"{names[moment]}={value:.3g}")

        seen.append(moment)

    return "\n" + ", ".join(str_m)


def get_slider(name, value, lower, upper):
    min_v, max_v, step = generate_range(value, lower, upper)

    if isinstance(value, float):
        slider_type = FloatSlider
    else:
        slider_type = IntSlider
        step = 1

    slider = slider_type(
        min=min_v,
        max=max_v,
        step=step,
        description=f"{name} ({lower:.0f}, {upper:.0f})",
        value=value,
        style={"description_width": "initial"},
    )

    return slider


def generate_range(value, lower, upper):
    if value == 0:
        return -10, 10, 0.1

    power = np.floor(np.log10(abs(value)))
    order_of_magnitude = 10**power
    decimal_places = int(-power)
    std = order_of_magnitude * 10 - order_of_magnitude / 10
    min_v = round(value - std, decimal_places)
    max_v = round(value + std, decimal_places)

    min_v = lower if np.isfinite(lower) else min_v
    max_v = upper if np.isfinite(upper) else max_v

    step = order_of_magnitude / 10
    return min_v, max_v, step


def get_boxes(name, value, lower, upper):
    if isinstance(value, float):
        text_type = FloatText
        step = 0.1
    else:
        text_type = IntText
        step = 1

    text = text_type(
        step=step,
        description=f"{name} ({lower:.0f}, {upper:.0f})",
        value=value,
        style={"description_width": "initial"},
    )

    return text


def get_textboxes(signature, model):
    textboxes = {}
    for name, param in signature.parameters.items():
        if isinstance(param.default, int | float):
            value = float(param.default)
        else:
            value = None

        dist, idx, func = model[name]
        lower, upper = dist.params_support[idx]
        # ((eps, 1 - eps),
        # (-np.inf, np.inf),
        # (eps, np.inf))
        # ((-np.pi, np.pi)
        if func is not None and lower == np.finfo(float).eps:
            if func in ["exp", "abs", "expit", "logistic"]:
                lower = -np.inf
            elif func in ["log"]:
                lower = 1.0
            elif func.replace(".", "", 1).isdigit():
                if not float(func) % 2:
                    lower = -np.inf

        if func is not None and upper == 1 - np.finfo(float).eps:
            if func in ["expit", "logistic"]:
                lower = np.inf

        if value is None:
            value = getattr(dist, dist.param_names[idx])

        textboxes[name] = get_boxes(name, value, lower, upper)

    textboxes["__set_xlim__"] = Checkbox(
        value=False, description="set xlim", disabled=False, indent=False
    )

    textboxes["__x_min__"] = FloatText(
        value=-10,
        step=0.1,
        description="x_min",
        disabled=False,
    )

    textboxes["__x_max__"] = FloatText(
        value=10,
        step=0.1,
        description="x_max",
        disabled=False,
    )

    textboxes["__set_ylim__"] = Checkbox(
        value=False, description="set ylim", disabled=False, indent=False
    )

    textboxes["__y_min__"] = FloatText(
        value=-10,
        step=0.1,
        description="y_min",
        disabled=False,
    )

    textboxes["__y_max__"] = FloatText(
        value=10,
        step=0.1,
        description="y_max",
        disabled=False,
    )

    textboxes["__resample__"] = ToggleButton(
        value=True,
        description="Resample",
        disabled=False,
        button_style="",
        tooltip="Resample",
    )

    return textboxes


def plot_decorator(func, iterations, kind_plot, references, plot_func):
    def looper(*args, **kwargs):
        results = []
        kwargs.pop("__resample__")
        x_min = kwargs.pop("__x_min__")
        x_max = kwargs.pop("__x_max__")
        if not kwargs.pop("__set_xlim__"):
            x_min = None
            x_max = None
            auto = True
        else:
            auto = False

        for _ in range(iterations):
            val = func(*args, **kwargs)
            if not any(np.isnan(val)):
                results.append(val)
        results = np.array(results)

        _, ax = plt.subplots()
        ax.set_xlim(x_min, x_max, auto=auto)

        if plot_func is None:
            plot_repr(results, kind_plot, references, iterations, ax)
        else:
            plot_func(results, ax)

    return looper


def plot_repr(results, kind_plot, references, iterations, ax):
    alpha = max(0.01, 1 - iterations * 0.009)

    if kind_plot == "hist":
        if results[0].dtype.kind == "i":
            bins = np.arange(np.min(results), np.max(results) + 1.5) - 0.5
            if len(bins) < 30:
                ax.set_xticks(bins + 0.5)
        else:
            bins = "auto"
        ax.hist(
            results.T,
            alpha=alpha,
            density=True,
            color=["0.5"] * iterations,
            bins=bins,
            histtype="step",
        )
        ax.hist(
            np.concatenate(results),
            density=True,
            bins=bins,
            color="k",
            ls="--",
            histtype="step",
        )
    elif kind_plot == "kde":
        for result in results:
            ax.plot(*kde(result), "0.5", alpha=alpha)
        ax.plot(*kde(np.concatenate(results)), "k--")
    elif kind_plot == "ecdf":
        ax.plot(
            np.sort(results, axis=1).T,
            np.linspace(0, 1, len(results[0]), endpoint=False),
            color="0.5",
        )
        a = np.concatenate(results)
        ax.plot(np.sort(a), np.linspace(0, 1, len(a), endpoint=False), "k--")

    plot_references(references, ax)


def plot_pp_samples(pp_samples, pp_samples_idxs, references, kind="pdf", sharex=True, fig=None):
    row_colum = int(np.ceil(len(pp_samples_idxs) ** 0.5))

    if fig is None:
        fig, axes = plt.subplots(row_colum, row_colum, figsize=(8, 6), constrained_layout=True)
        fig.canvas.header_visible = False
        fig.canvas.footer_visible = False
    else:
        axes = np.array(fig.axes)

    try:
        axes = axes.ravel()
    except AttributeError:
        axes = [axes]

    x_lims = [np.inf, -np.inf]

    for ax, idx in zip(axes, pp_samples_idxs):
        ax.clear()
        ax.relim()

        sample = pp_samples[idx]

        if kind == "pdf":
            plot_kde(sample, ax=ax, color="C0")
        elif kind == "hist":
            bins, *_ = ax.hist(sample, color="C0", bins="auto", alpha=0.5, density=True)
            ax.set_ylim(-bins.max() * 0.05, None)

        elif kind == "ecdf":
            ax.ecdf(sample, color="C0")

        plot_pointinterval(sample, ax=ax)
        plot_references(references, ax)

        if sharex:
            min_, max_ = ax.get_xlim()
            x_lims[0] = min(min_, x_lims[0])
            x_lims[1] = max(max_, x_lims[1])

        ax.set_title(idx, alpha=0)
        ax.set_yticks([])

    if sharex:
        for ax in axes:
            ax.set_xlim(np.floor(x_lims[0]), np.ceil(x_lims[1]))

    fig.canvas.draw()
    return fig, axes


def plot_pp_mean(pp_samples, selected, references=None, kind="pdf", fig_pp_mean=None):
    if fig_pp_mean is None:
        fig_pp_mean, ax_pp_mean = plt.subplots(1, 1, figsize=(8, 2), constrained_layout=True)
        fig_pp_mean.canvas.header_visible = False
        fig_pp_mean.canvas.footer_visible = False
    else:
        ax_pp_mean = fig_pp_mean.axes[0]

    ax_pp_mean.clear()
    ax_pp_mean.relim()

    if np.any(selected):
        sample = pp_samples[selected].ravel()
    else:
        sample = pp_samples.ravel()

    if kind == "pdf":
        plot_kde(
            sample,
            ax=ax_pp_mean,
            color="C0",
            linestyle="--",
        )
    elif kind == "hist":
        bins, *_ = ax_pp_mean.hist(sample, color="k", ls="--", bins="auto", alpha=0.5, density=True)
        ax_pp_mean.set_ylim(-bins.max() * 0.05, None)

    elif kind == "ecdf":
        ax_pp_mean.ecdf(sample, color="k", linestyle="--")

    plot_pointinterval(sample, ax=ax_pp_mean)
    plot_references(references, ax_pp_mean)
    ax_pp_mean.set_yticks([])
    fig_pp_mean.canvas.draw()

    return fig_pp_mean


def plot_references(references, ax):
    if references is not None:
        if isinstance(references, dict):
            max_value = ax.get_ylim()[1]
            for label, ref in references.items():
                ax.text(ref, max_value * 0.2, label, rotation=90, bbox={"color": "w", "alpha": 0.5})
                ax.axvline(ref, ls="--", color="0.5")
        else:
            if isinstance(references, float | int):
                references = [references]
            for ref in references:
                ax.axvline(ref, ls="--", color="0.5")


def plot_kde(sample, ax, **kwargs):
    grid, pdf = kde(sample)
    ax.plot(grid, pdf, **kwargs)


def check_inside_notebook(need_widget=False):
    shell = get_ipython()
    name = inspect.currentframe().f_back.f_code.co_name
    try:
        if shell is None:
            raise RuntimeError(
                f"To run {name}, you need to call it from within a Jupyter notebook or Jupyter lab."
            )
        if need_widget:
            shell_name = shell.__class__.__name__
            if shell_name == "ZMQInteractiveShell" and get_backend().split("_")[-1].lower() not in [
                "ipympl",
                "widget",
                "nbagg",
            ]:
                msg = f"To run {name}, you need use the magic `%matplotlib widget`"
                raise RuntimeError(msg)
    except Exception:
        tb_as_str = traceback.format_exc()
        # Print only the last line of the traceback, which contains the error message
        print(tb_as_str.strip().rsplit("\n", maxsplit=1)[-1], file=sys.stdout)


def representations(fitted_dist, kind_plot, ax):
    if kind_plot == "pdf":
        fitted_dist.plot_pdf(pointinterval=True, legend="title", ax=ax)
        ax.set_yticks([])

        for bound in fitted_dist.support:
            if np.isfinite(bound):
                ax.plot(bound, 0, "ko")

    elif kind_plot == "cdf":
        fitted_dist.plot_cdf(pointinterval=True, legend="title", ax=ax)

    elif kind_plot == "ppf":
        fitted_dist.plot_ppf(pointinterval=True, legend="title", ax=ax)
        ax.set_xlim(-0.01, 1)


def create_figure(figsize):
    """Initialize a matplotlib figure with one subplot."""
    fig, axes = plt.subplots(1, 1, figsize=figsize, constrained_layout=True)
    axes.set_yticks([])
    fig.canvas.header_visible = False
    fig.canvas.footer_visible = False
    fig.canvas.toolbar_position = "right"

    return fig, axes


def reset_dist_panel(ax, yticks):
    """Clean the distribution subplot."""
    ax.cla()
    if yticks:
        ax.set_yticks([])
    ax.relim()
    ax.autoscale_view()


def set_label(dist, legend, moments, ax):
    if legend is not None:
        if isinstance(legend, str) and legend not in ["title", "legend"]:
            label = legend
        else:
            label = repr_to_matplotlib(dist)

            if moments is not None:
                label += get_moments(dist, moments)

            if legend == "title":
                ax.set_title(label)
                label = None
    else:
        label = None
    return label



import pytest
from numpy.testing import assert_almost_equal

from preliz import combine_roulette
from preliz.distributions import BetaScaled, LogNormal, StudentT

response0 = (
    [1.5, 2.5, 3.5],
    [0.32142857142857145, 0.35714285714285715, 0.32142857142857145],
    28,
    0,
    10,
    10,
    11,
)
response1 = (
    [7.5, 8.5, 9.5],
    [0.32142857142857145, 0.35714285714285715, 0.32142857142857145],
    28,
    0,
    10,
    10,
    11,
)
response2 = ([9.5], [1], 10, 0, 10, 10, 11)
response3 = ([9.5], [1], 10, 0, 10, 10, 14)


@pytest.mark.parametrize(
    "responses, weights, dist_names, params, result",
    [
        ([response0, response1], [0.5, 0.5], None, None, BetaScaled(1.2, 1, 0, 10)),
        (
            [response0, response1],
            [0.5, 0.5],
            ["Beta", "StudentT"],
            "TruncatedNormal(lower=0), StudentT(nu=1000)",
            StudentT(1000, 5.5, 3.1),
        ),
        ([response0, response2], [1, 1], None, None, LogNormal(1.1, 0.6)),
    ],
)
def test_combine_roulette(responses, weights, dist_names, params, result):
    dist = combine_roulette(responses, weights, dist_names, params)
    assert_almost_equal(dist.params, result.params, decimal=1)


def test_combine_roulette_error():
    with pytest.raises(ValueError):
        combine_roulette([response0, response3])



import numba as nb
import numpy as np
from scipy.special import erf, erfinv

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.special import (
    cdf_bounds,
    erf,  # noqa: F811
    erfinv,  # noqa: F811
    mean_and_std,
    ppf_bounds_cont,
)


class LogNormal(Continuous):
    r"""
    Log-normal distribution.

    Distribution of any random variable whose logarithm is normally
    distributed. A variable might be modeled as log-normal if it can
    be thought of as the multiplicative product of many small
    independent factors.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu, \sigma) =
          \frac{1}{x \sigma \sqrt{2\pi}}
           \exp\left\{ -\frac{(\ln(x)-\mu)^2}{2\sigma^2} \right\}

    .. plot::
        :context: close-figs


        from preliz import LogNormal, style
        style.use('preliz-doc')
        mus = [ 0., 0.]
        sigmas = [.5, 1.]
        for mu, sigma in zip(mus, sigmas):
            LogNormal(mu, sigma).plot_pdf(support=(0,5))

    ========  =========================================================================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\exp\left(\mu+\frac{\sigma^2}{2}\right)`
    Variance  :math:`[\exp(\sigma^2)-1] \exp(2\mu+\sigma^2)`
    ========  =========================================================================

    Parameters
    ----------
    mu : float
        Location parameter.
    sigma : float
        Standard deviation. (sigma > 0)).
    """

    def __init__(self, mu=None, sigma=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(mu, sigma)

    def _parametrization(self, mu=None, sigma=None):
        self.mu = mu
        self.sigma = sigma
        self.params = (self.mu, self.sigma)
        self.param_names = ("mu", "sigma")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        if all_not_none(mu, sigma):
            self._update(mu, sigma)

    def _update(self, mu, sigma):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.params = (self.mu, self.sigma)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma)

    def entropy(self):
        return nb_entropy(self.mu, self.sigma)

    def mean(self):
        return np.exp(self.mu + self.sigma**2 / 2)

    def mode(self):
        return np.exp(self.mu - self.sigma**2)

    def median(self):
        return np.exp(self.mu)

    def var(self):
        return (np.exp(self.sigma**2) - 1) * np.exp(2 * self.mu + self.sigma**2)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (np.exp(self.sigma**2) + 2) * (np.exp(self.sigma**2) - 1) ** 0.5

    def kurtosis(self):
        return (
            np.exp(4 * self.sigma**2)
            + 2 * np.exp(3 * self.sigma**2)
            + 3 * np.exp(2 * self.sigma**2)
            - 6
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.lognormal(self.mu, self.sigma, size)

    def _fit_moments(self, mean, sigma):
        mu = np.log(mean**2 / (sigma**2 + mean**2) ** 0.5)
        sigma = np.log(sigma**2 / mean**2 + 1) ** 0.5
        self._update(mu, sigma)

    def _fit_mle(self, sample):
        mu, sigma = mean_and_std(np.log(sample))
        self._update(mu, sigma)


@nb.njit(cache=True)
def nb_cdf(x, mu, sigma):
    return cdf_bounds(0.5 * (1 + erf((np.log(x) - mu) / (sigma * 2**0.5))), x, 0, np.inf)


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma):
    return ppf_bounds_cont(np.exp(mu + sigma * 2**0.5 * erfinv(2 * q - 1)), q, 0, np.inf)


@nb.njit(cache=True)
def nb_entropy(mu, sigma):
    return np.log((2 * np.pi) ** 0.5 * sigma * np.exp(mu + 0.5))


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, sigma):
    if x <= 0:
        return -np.inf
    else:
        return -(np.log(x) + np.log(sigma) + 0.5 * np.log(2 * np.pi)) - ((np.log(x) - mu) ** 2) / (
            2 * sigma**2
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma):
    return -(nb_logpdf(x, mu, sigma)).sum()



import numba as nb
import numpy as np
from scipy.stats import skew

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import find_ppf
from preliz.internal.special import erf, mean_and_std, norm_logcdf


class ExGaussian(Continuous):
    r"""
    Exponentially modified Gaussian (EMG) Distribution.

    Results from the convolution of a normal distribution with an exponential
    distribution.

    The pdf of this distribution is

    .. math::

        f(x \mid \mu, \sigma, \nu) =
            \frac{1}{\nu}\;
            \exp\left\{\frac{\mu-x}{\nu}+\frac{\sigma^2}{2\nu^2}\right\}
            \Phi\left(\frac{x-\mu}{\sigma}-\frac{\sigma}{\nu}\right)

    where :math:`\Phi` is the cumulative distribution function of the
    standard normal distribution.

    .. plot::
        :context: close-figs


        from preliz import ExGaussian, style
        style.use('preliz-doc')
        mus = [0., 0., -3.]
        sigmas = [1., 3., 1.]
        nus = [1., 1., 4.]
        for mu, sigma, nu in zip(mus, sigmas, nus):
            ExGaussian(mu, sigma, nu).plot_pdf(support=(-6,9))

    ========  ========================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu + \nu`
    Variance  :math:`\sigma^2 + \nu^2`
    ========  ========================

    Parameters
    ----------
    mu : float
        Mean of the normal distribution.
    sigma : float
        Standard deviation of the normal distribution (sigma > 0).
    nu : float
        Mean of the exponential distribution (nu > 0).
    """

    def __init__(self, mu=None, sigma=None, nu=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, sigma, nu)

    def _parametrization(self, mu=None, sigma=None, nu=None):
        self.mu = mu
        self.sigma = sigma
        self.nu = nu
        self.param_names = ("mu", "sigma", "nu")
        self.params = (mu, sigma, nu)
        #  if nu is too small we get a non-smooth distribution
        self.params_support = ((-np.inf, np.inf), (eps, np.inf), (1e-4, np.inf))
        if all_not_none(mu, sigma, nu):
            self._update(mu, sigma, nu)

    def _update(self, mu, sigma, nu):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.nu = np.float64(nu)
        self.params = (self.mu, self.sigma, self.nu)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma, self.nu)

    def ppf(self, q):
        q = np.asarray(q)
        return find_ppf(self, q)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma, self.nu)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma, self.nu)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return self.mu + self.nu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.sigma**2 + self.nu**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        nus2 = (self.nu / self.sigma) ** 2
        opnus2 = 1.0 + nus2
        return 2 * (self.nu / self.sigma) ** 3 * opnus2 ** (-1.5)

    def kurtosis(self):
        nus2 = (self.nu / self.sigma) ** 2
        opnus2 = 1.0 + nus2
        return 6.0 * nus2 * nus2 * opnus2 ** (-2)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.normal(self.mu, self.sigma, size) + random_state.exponential(
            self.nu, size
        )

    def _fit_moments(self, mean, sigma):
        # Just assume this is a approximately Gaussian
        self._update(mean, sigma, 1e-4)

    def _fit_mle(self, sample):
        mean, std = mean_and_std(sample)
        skweness = max(1e-4, skew(sample))
        nu = std * (skweness / 2) ** (1 / 3)
        mu = mean - nu
        var = std**2 * (1 - (skweness / 2) ** (2 / 3))
        self._update(mu, var**0.5, nu)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, mu, sigma, nu):
    cdf_n = 0.5 * (1 + erf((x - mu) / (sigma * 2**0.5)))
    if x == -np.inf:
        return 0
    elif nu > 0.05 * sigma:
        return cdf_n - 0.5 * np.exp(0.5 / nu * (2 * mu + sigma**2 / nu - 2 * x)) * (
            1 + erf((x - (mu + (sigma**2) / nu)) / (sigma * 2**0.5))
        )
    else:
        return cdf_n


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, sigma, nu):
    if nu > 0.05 * sigma:
        return (
            -np.log(nu)
            + (mu - x) / nu
            + 0.5 * (sigma / nu) ** 2
            + norm_logcdf((x - (mu + (sigma**2) / nu)) / sigma)
        )
    else:
        return -np.log(sigma) - 0.5 * np.log(2 * np.pi) - 0.5 * ((x - mu) / sigma) ** 2


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma, nu):
    return -(nb_logpdf(x, mu, sigma, nu)).sum()



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from preliz.distributions import (
    Beta,
    Exponential,
    Geometric,
    HalfNormal,
    Laplace,
    Normal,
    Poisson,
    StudentT,
    Weibull,
)
from preliz.internal.optimization import find_ppf


@pytest.mark.parametrize(
    "p_dist, p_params",
    [
        (Beta, {"alpha": 2, "beta": 5}),
        (Exponential, {"beta": 3.7}),
        (HalfNormal, {"sigma": 2}),
        (Laplace, {"mu": 2.5, "b": 4}),
        (Normal, {"mu": 0, "sigma": 2}),
        (StudentT, {"nu": 5, "mu": 0, "sigma": 2}),
        (Weibull, {"alpha": 5.0, "beta": 2.0}),
        (Geometric, {"p": 0.4}),
        (Poisson, {"mu": 3.5}),
    ],
)
def test_find_ppf(p_dist, p_params):
    preliz_dist = p_dist(**p_params)
    x_vals = np.linspace(0.001, 0.999, 10)
    actual_ppf = preliz_dist.ppf(x_vals)
    expected_ppf = find_ppf(preliz_dist, x_vals)
    assert_almost_equal(actual_ppf, expected_ppf, decimal=4)
